@@ -382,8 +382,8 @@ static SYS_GETXATTRAT: Lazy<libc::c_long> = Lazy::new(|| {
 /// This function dereferences `name` argument which is a raw pointer.
 /// It is safe for this pointer to be NULL but otherwise it must point
 /// to valid memory and the caller must ensure that.
-pub unsafe fn getxattrat<F: AsRawFd, P: ?Sized + NixPath>(
-    dirfd: Option<&F>,
+pub unsafe fn getxattrat<Fd: AsFd, P: ?Sized + NixPath>(
+    dirfd: Fd,
     path: &P,
     name: *const libc::c_char,
     args: &mut XattrArgs,
@@ -403,7 +403,7 @@ pub unsafe fn getxattrat<F: AsRawFd, P: ?Sized + NixPath>(
             unsafe {
                 libc::syscall(
                     sysno as libc::c_long,
-                    dirfd.map(|fd| fd.as_raw_fd()).unwrap_or(libc::AT_FDCWD),
+                    dirfd.as_fd().as_raw_fd(),
                     c_path.as_ptr(),
                     flags.bits(),
                     name,
@@ -433,8 +433,8 @@ static SYS_SETXATTRAT: Lazy<libc::c_long> = Lazy::new(|| {
 /// This function dereferences `name` argument which is a raw pointer.
 /// It is safe for this pointer to be NULL but otherwise it must point
 /// to valid memory and the caller must ensure that.
-pub unsafe fn setxattrat<F: AsRawFd, P: ?Sized + NixPath>(
-    dirfd: Option<&F>,
+pub unsafe fn setxattrat<Fd: AsFd, P: ?Sized + NixPath>(
+    dirfd: Fd,
     path: &P,
     name: *const libc::c_char,
     args: &XattrArgs,
@@ -451,7 +451,7 @@ pub unsafe fn setxattrat<F: AsRawFd, P: ?Sized + NixPath>(
         Errno::result(unsafe {
             libc::syscall(
                 sysno as libc::c_long,
-                dirfd.map(|fd| fd.as_raw_fd()).unwrap_or(libc::AT_FDCWD),
+                dirfd.as_fd().as_raw_fd(),
                 c_path.as_ptr(),
                 flags.bits(),
                 name,
@@ -480,8 +480,8 @@ static SYS_LISTXATTRAT: Lazy<libc::c_long> = Lazy::new(|| {
 /// This function dereferences `addr` argument which is a raw pointer.
 /// It is safe for this pointer to be NULL but otherwise it must point
 /// to valid memory and the caller must ensure that.
-pub unsafe fn listxattrat<F: AsRawFd, P: ?Sized + NixPath>(
-    dirfd: Option<&F>,
+pub unsafe fn listxattrat<Fd: AsFd, P: ?Sized + NixPath>(
+    dirfd: Fd,
     path: &P,
     flags: AtFlags,
     addr: *mut libc::c_char,
@@ -501,7 +501,7 @@ pub unsafe fn listxattrat<F: AsRawFd, P: ?Sized + NixPath>(
             unsafe {
                 libc::syscall(
                     sysno as libc::c_long,
-                    dirfd.map(|fd| fd.as_raw_fd()).unwrap_or(libc::AT_FDCWD),
+                    dirfd.as_fd().as_raw_fd(),
                     c_path.as_ptr(),
                     flags.bits(),
                     addr,
@@ -530,8 +530,8 @@ static SYS_REMOVEXATTRAT: Lazy<libc::c_long> = Lazy::new(|| {
 /// This function dereferences `name` argument which is a raw pointer.
 /// It is safe for this pointer to be NULL but otherwise it must point
 /// to valid memory and the caller must ensure that.
-pub unsafe fn removexattrat<F: AsRawFd, P: ?Sized + NixPath>(
-    dirfd: Option<&F>,
+pub unsafe fn removexattrat<Fd: AsFd, P: ?Sized + NixPath>(
+    dirfd: Fd,
     path: &P,
     name: *const libc::c_char,
     flags: AtFlags,
@@ -547,7 +547,7 @@ pub unsafe fn removexattrat<F: AsRawFd, P: ?Sized + NixPath>(
         Errno::result(unsafe {
             libc::syscall(
                 sysno as libc::c_long,
-                dirfd.map(|fd| fd.as_raw_fd()).unwrap_or(libc::AT_FDCWD),
+                dirfd.as_fd().as_raw_fd(),
                 c_path.as_ptr(),
                 flags.bits(),
                 name,
@@ -742,11 +742,11 @@ impl DirIter {
     ///
     /// Returns `Errno` if syscall fails or EOF reached.
     /// EOF is indicated by `Errno::ECANCELED`.
-    pub fn readdir<F: AsRawFd>(&mut self, fd: &F, read_bufsiz: usize) -> Result<&mut Self, Errno> {
+    pub fn readdir<Fd: AsFd>(&mut self, fd: Fd, read_bufsiz: usize) -> Result<&mut Self, Errno> {
         self.offset = 0;
         let bufsiz = read_bufsiz.min(self.memsiz);
 
-        let retsiz = sys_getdents64(fd.as_raw_fd(), self.buffer.as_ptr().cast(), bufsiz)?;
+        let retsiz = sys_getdents64(fd, self.buffer.as_ptr().cast(), bufsiz)?;
         if retsiz == 0 {
             return Err(Errno::ECANCELED); // EOF or empty directory
         }
@@ -947,7 +947,7 @@ impl DirEntry<'_> {
 ///
 /// - `fd`: The open directory file descriptor.
 /// - `bufsiz`: The maximum number of bytes to read into the buffer.
-///             If greater than internal buffer size, internal size is used.
+///   If greater than internal buffer size, internal size is used.
 ///
 /// # Returns
 ///
@@ -958,8 +958,8 @@ impl DirEntry<'_> {
 ///
 /// Internally calls the unsafe _getdents64_(2) syscall.
 /// The provided file descriptor must be a valid, open directory descriptor.
-pub fn getdents64<F: AsRawFd>(
-    fd: &F,
+pub fn getdents64<Fd: AsFd>(
+    fd: Fd,
     bufsiz: usize,
 ) -> Result<impl Iterator<Item = DirEntry<'static>>, Errno> {
     thread_local! {
@@ -997,12 +997,12 @@ pub fn getdents64<F: AsRawFd>(
 /// Wrapper for the `getdents64` syscall.
 #[allow(clippy::cast_possible_truncation)]
 #[allow(clippy::cast_sign_loss)]
-fn sys_getdents64(fd: RawFd, buf: *mut libc::c_void, bytes: usize) -> Result<usize, Errno> {
-    Ok(
-        // SAFETY: In kernel, we trust.
-        Errno::result(unsafe { libc::syscall(nix::libc::SYS_getdents64, fd, buf, bytes) })?
-            as usize,
-    )
+fn sys_getdents64<Fd: AsFd>(fd: Fd, buf: *mut libc::c_void, bytes: usize) -> Result<usize, Errno> {
+    // SAFETY: In kernel, we trust.
+    Errno::result(unsafe {
+        libc::syscall(libc::SYS_getdents64, fd.as_fd().as_raw_fd(), buf, bytes)
+    })
+    .map(|size| size as usize)
 }
 
 /// WaitStatus with support for signals that nix' Signal type don't support.
@@ -1107,9 +1107,9 @@ pub fn waitid(id: Id, flags: WaitPidFlag) -> Result<WaitStatus, Errno> {
     #[allow(clippy::cast_sign_loss)]
     let (idtype, idval) = match id {
         Id::All => (libc::P_ALL, 0),
-        Id::Pid(pid) => (libc::P_PID, pid.as_raw() as nix::libc::id_t),
-        Id::PGid(pid) => (libc::P_PGID, pid.as_raw() as nix::libc::id_t),
-        Id::PIDFd(fd) => (libc::P_PIDFD, fd.as_raw_fd() as nix::libc::id_t),
+        Id::Pid(pid) => (libc::P_PID, pid.as_raw() as libc::id_t),
+        Id::PGid(pid) => (libc::P_PGID, pid.as_raw() as libc::id_t),
+        Id::PIDFd(fd) => (libc::P_PIDFD, fd.as_raw_fd() as libc::id_t),
         _ => unreachable!(),
     };
 
@@ -1136,7 +1136,7 @@ pub fn waitid(id: Id, flags: WaitPidFlag) -> Result<WaitStatus, Errno> {
 
     let status = match siginfo.si_code {
         libc::CLD_EXITED => WaitStatus::Exited(pid, si_status),
-        libc::CLD_KILLED | nix::libc::CLD_DUMPED => {
+        libc::CLD_KILLED | libc::CLD_DUMPED => {
             WaitStatus::Signaled(pid, si_status, siginfo.si_code == libc::CLD_DUMPED)
         }
         libc::CLD_STOPPED => WaitStatus::Stopped(pid, si_status),
@@ -1170,17 +1170,17 @@ pub(crate) fn pipe2_raw(flags: OFlag) -> Result<(RawFd, RawFd), Errno> {
 }
 
 #[allow(clippy::cast_possible_truncation)]
-pub(crate) const PF_UNSPEC: libc::sa_family_t = nix::libc::AF_UNSPEC as nix::libc::sa_family_t;
+pub(crate) const PF_UNSPEC: libc::sa_family_t = libc::AF_UNSPEC as libc::sa_family_t;
 #[allow(clippy::cast_possible_truncation)]
-pub(crate) const PF_UNIX: libc::sa_family_t = nix::libc::AF_UNIX as nix::libc::sa_family_t;
+pub(crate) const PF_UNIX: libc::sa_family_t = libc::AF_UNIX as libc::sa_family_t;
 #[allow(clippy::cast_possible_truncation)]
-pub(crate) const PF_INET: libc::sa_family_t = nix::libc::AF_INET as nix::libc::sa_family_t;
+pub(crate) const PF_INET: libc::sa_family_t = libc::AF_INET as libc::sa_family_t;
 #[allow(clippy::cast_possible_truncation)]
-pub(crate) const PF_INET6: libc::sa_family_t = nix::libc::AF_INET6 as nix::libc::sa_family_t;
+pub(crate) const PF_INET6: libc::sa_family_t = libc::AF_INET6 as libc::sa_family_t;
 #[allow(clippy::cast_possible_truncation)]
-pub(crate) const PF_ALG: libc::sa_family_t = nix::libc::AF_ALG as nix::libc::sa_family_t;
+pub(crate) const PF_ALG: libc::sa_family_t = libc::AF_ALG as libc::sa_family_t;
 #[allow(clippy::cast_possible_truncation)]
-pub(crate) const PF_NETLINK: libc::sa_family_t = libc::AF_NETLINK as nix::libc::sa_family_t;
+pub(crate) const PF_NETLINK: libc::sa_family_t = libc::AF_NETLINK as libc::sa_family_t;
 
 /// nix' SockAddrLike.family() function does not support AF_ALG.
 /// This is a workaround that accesses the underlying sockaddr directly.
@@ -1235,13 +1235,13 @@ pub const AT_STATX_FORCE_SYNC: libc::c_int = 0x2000;
 pub const AT_STATX_DONT_SYNC: libc::c_int = 0x4000;
 
 /// Safe statx() wrapper.
-pub fn statx<F: AsRawFd, P: ?Sized + NixPath>(
-    dirfd: Option<&F>,
+pub fn statx<Fd: AsFd, P: ?Sized + NixPath>(
+    dirfd: Fd,
     pathname: &P,
     flags: libc::c_int,
     mask: libc::c_uint,
 ) -> Result<FileStatx, Errno> {
-    let dirfd = dirfd.map(|fd| fd.as_raw_fd()).unwrap_or(libc::AT_FDCWD);
+    let dirfd = dirfd.as_fd().as_raw_fd();
     let mut dst = std::mem::MaybeUninit::uninit();
 
     // SAFETY: Neither nix nor libc has a wrapper for statx.
@@ -1261,8 +1261,8 @@ pub fn statx<F: AsRawFd, P: ?Sized + NixPath>(
 }
 
 /// Safe statx() wrapper to use with a FD only.
-pub fn fstatx<F: AsRawFd>(fd: &F, mask: libc::c_uint) -> Result<FileStatx, Errno> {
-    let fd = fd.as_raw_fd();
+pub fn fstatx<Fd: AsFd>(fd: Fd, mask: libc::c_uint) -> Result<FileStatx, Errno> {
+    let fd = fd.as_fd().as_raw_fd();
     let mut dst = std::mem::MaybeUninit::uninit();
 
     // SAFETY: Neither nix nor libc has a wrapper for statx.
@@ -1345,11 +1345,11 @@ impl Statfs64 {
 }
 
 /// Safe wrapper for fstatfs64
-pub(crate) fn fstatfs64<F: AsRawFd>(fd: &F) -> Result<Statfs64, Errno> {
+pub(crate) fn fstatfs64<Fd: AsFd>(fd: Fd) -> Result<Statfs64, Errno> {
     let mut dst = std::mem::MaybeUninit::uninit();
 
     // SAFETY: nix does not have a wrapper for fstatfs64.
-    Errno::result(unsafe { libc::fstatfs64(fd.as_raw_fd(), dst.as_mut_ptr()) })?;
+    Errno::result(unsafe { libc::fstatfs64(fd.as_fd().as_raw_fd(), dst.as_mut_ptr()) })?;
 
     // SAFETY: fstatfs64 returned success.
     Ok(Statfs64(unsafe { dst.assume_init() }))
@@ -1449,33 +1449,35 @@ impl Serialize for EpollParams {
 
 /// Set epoll parameters for the given epoll file descriptor.
 /// Requires Linux>=6.9.
-pub fn epoll_set_params<F: AsRawFd>(fd: &F, params: &EpollParams) -> Result<(), Errno> {
+pub fn epoll_set_params<Fd: AsFd>(fd: Fd, params: &EpollParams) -> Result<(), Errno> {
     // SAFETY: In libc we trust.
-    Errno::result(unsafe { libc::ioctl(fd.as_raw_fd(), EPIOCSPARAMS, params) }).map(drop)
+    Errno::result(unsafe { libc::ioctl(fd.as_fd().as_raw_fd(), EPIOCSPARAMS, params) }).map(drop)
 }
 
 /// Get epoll parameters for the given epoll file descriptor.
 /// Requires Linux>=6.9.
-pub fn epoll_get_params<F: AsRawFd>(fd: &F) -> Result<EpollParams, Errno> {
+pub fn epoll_get_params<Fd: AsFd>(fd: Fd) -> Result<EpollParams, Errno> {
     let mut params = std::mem::MaybeUninit::uninit();
 
     // SAFETY: In libc we trust.
-    Errno::result(unsafe { libc::ioctl(fd.as_raw_fd(), EPIOCGPARAMS, params.as_mut_ptr()) })?;
+    Errno::result(unsafe {
+        libc::ioctl(fd.as_fd().as_raw_fd(), EPIOCGPARAMS, params.as_mut_ptr())
+    })?;
 
     // SAFETY: ioctl returned success.
     Ok(unsafe { params.assume_init() })
 }
 
 /// Uses getsockopt SO_DOMAIN to get the domain of the given socket.
-pub fn getsockdomain<F: AsRawFd>(fd: &F) -> Result<libc::c_int, Errno> {
+pub fn getsockdomain<Fd: AsFd>(fd: Fd) -> Result<libc::c_int, Errno> {
     #[allow(clippy::cast_possible_truncation)]
-    let mut len = std::mem::size_of::<libc::c_int>() as nix::libc::socklen_t;
+    let mut len = std::mem::size_of::<libc::c_int>() as libc::socklen_t;
     let mut fml: libc::c_int = 0;
 
     // SAFETY: In libc we trust.
     Errno::result(unsafe {
         libc::getsockopt(
-            fd.as_raw_fd(),
+            fd.as_fd().as_raw_fd(),
             libc::SOL_SOCKET,
             libc::SO_DOMAIN,
             std::ptr::addr_of_mut!(fml) as *mut _,
@@ -1575,7 +1577,7 @@ pub fn lsm_list_modules() -> Result<Vec<LsmId>, Errno> {
             libc::syscall(
                 sysno as libc::c_long,
                 std::ptr::null_mut::<u64>(),
-                &mut size as *mut u32,
+                std::ptr::addr_of_mut!(size),
                 0u32,
             )
         },
@@ -1595,13 +1597,15 @@ pub fn lsm_list_modules() -> Result<Vec<LsmId>, Errno> {
 
     let count = (size / 8) as usize;
     let mut buf = vec![0u64; count];
+    #[allow(clippy::cast_possible_truncation)]
+    #[allow(clippy::cast_sign_loss)]
     let count = Errno::result(
         // SAFETY: Properly allocated buffer & valid pointer to `size`.
         unsafe {
             libc::syscall(
                 sysno as libc::c_long,
                 buf.as_mut_ptr(),
-                &mut size as *mut u32,
+                std::ptr::addr_of_mut!(size),
                 0u32,
             )
         },
@@ -1615,8 +1619,8 @@ pub fn lsm_list_modules() -> Result<Vec<LsmId>, Errno> {
 
     // Populate output array.
     let mut out = Vec::with_capacity(count);
-    for i in 0..count {
-        out.push(LsmId::from(buf[i]));
+    for item in buf.iter().take(count).copied().map(LsmId::from) {
+        out.push(item)
     }
     Ok(out)
 }
