@@ -21,7 +21,7 @@
 
 use std::{
     os::{
-        fd::{BorrowedFd, RawFd},
+        fd::{AsRawFd, BorrowedFd, RawFd},
         unix::ffi::OsStrExt,
     },
     sync::OnceLock,
@@ -40,9 +40,9 @@ use once_cell::sync::Lazy;
 
 use crate::{
     compat::{STATX_MNT_ID, STATX_MNT_ID_UNIQUE},
-    fs::duprand,
+    fs::{duprand, FileMap, FileMapEntry, FileType},
     fstatx, info,
-    path::XPath,
+    path::{XPath, XPathBuf},
     pool::PidFdMap,
     proc::proc_mmap_min_addr,
     sysinfo::RandTimer,
@@ -79,10 +79,6 @@ pub const ENV_NPROC: &str = "SYD_NPROC";
 pub const ENV_NPROC_MAX: &str = "SYD_NPROC_MAX";
 /// The environment variable to read the pid filename from.
 pub const ENV_PID_FN: &str = "SYD_PID_FN";
-/// The environment variable to set to the poll file descriptor.
-///
-/// This variable is used internally and must not be set manually.
-pub const ENV_POLL_FD: &str = "SYD_POLL_FD";
 /// The environment variable to force TTY output.
 pub const ENV_FORCE_TTY: &str = "SYD_FORCE_TTY";
 /// The environment variable to quiet TTY output.
@@ -92,6 +88,8 @@ pub const ENV_QUIET_TTY: &str = "SYD_QUIET_TTY";
 pub const ENV_PROXY_HOST: &str = "SYD_PROXY_HOST";
 /// The environment variable to read the default external port from.
 pub const ENV_PROXY_PORT: &str = "SYD_PROXY_PORT";
+/// The environment variable to read the default external unix socket from.
+pub const ENV_PROXY_UNIX: &str = "SYD_PROXY_UNIX";
 
 /// The environment variable to set to disable logging to syslog.
 pub const ENV_NO_SYSLOG: &str = "SYD_NO_SYSLOG";
@@ -441,21 +439,6 @@ pub const UNSAFE_ENV: &[&[u8]] = &[
     b"TZDIR",
 ];
 
-/// Path prefixes to skip caching.
-///
-/// Note, trailing slash matters!
-pub const SKIP_CACHE_PATH_PREFIXES: &[&[u8]] = &[
-    b"/boot",
-    b"/root",
-    b"/dev/", // we pre-open `/dev`.
-    b"/etc",
-    b"/proc/", // we pre-open `/proc`.
-    b"/run",
-    b"/sys/", // we pre-open `/sys`.
-    b"/var/log",
-    b"/var/run",
-];
-
 /// Default ioctl allow/denylist.
 /// false -> allow, true -> deny.
 pub const DEFAULT_IOCTL: &[(u64, bool)] = &[
@@ -563,6 +546,10 @@ pub const PROFILE_NOMEM: &[&str] = &["trace/allow_unsafe_memory:true"];
 /// No PIE sandboxing profile.
 pub const PROFILE_NOPIE: &[&str] = &["trace/allow_unsafe_nopie:true"];
 
+/// Readonly sandboxing profile.
+/// `syd --profile ro`
+pub const PROFILE_READONLY: &[&str] = &["deny/wrset+/***"];
+
 /// Container sandboxing profile, activated with:
 /// `syd --profile container`
 pub const PROFILE_CONTAINER: &[&str] = &["unshare/all:true"];
@@ -575,47 +562,12 @@ pub const PROFILE_IMMUTABLE: &[&str] = &[
     "bind+tmpfs:/dev/shm:nodev,nosuid,noexec,nosymfollow,mode=1777",
     "bind+tmpfs:/run:nodev,nosuid,nosymfollow,mode=1777",
     "bind+tmpfs:/tmp:nodev,nosuid,nosymfollow,mode=1777",
-    "bind+/run:/var/run:nodev,nosuid,nosymfollow,mode=1777",
     "bind+/etc:/etc:ro,nodev,noexec,nosuid",
     "bind+/home:/home:ro,nodev,noexec,nosuid",
-    "bind+/media:/media:ro,nodev,noexec,nosuid",
     "bind+/mnt:/mnt:ro,nodev,noexec,nosuid",
     "bind+/opt:/opt:ro,nodev,nosuid",
     "bind+/srv:/srv:ro,nodev,noexec,nosuid",
     "bind+/usr:/usr:ro,nodev,nosuid",
-    "bind+/etc/shells:/proc/kcore:ro,nodev,noexec,nosuid",
-    "bind+/etc/shells:/proc/keys:ro,nodev,noexec,nosuid",
-    "bind+/etc/shells:/proc/latency_stats:ro,nodev,noexec,nosuid",
-    "bind+/etc/shells:/proc/sys/kernel/modprobe:ro,nodev,noexec,nosuid",
-    "bind+/etc/shells:/proc/sysrq-trigger:ro,nodev,noexec,nosuid",
-    "bind+/etc/shells:/proc/timer_list:ro,nodev,noexec,nosuid",
-    "bind+/etc/shells:/proc/timer_stats:ro,nodev,noexec,nosuid",
-    "bind+/etc/shells:/sys/kernel/notes:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/proc/acpi:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/proc/asound:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/proc/bus:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/proc/driver:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/proc/dynamic_debug:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/proc/fs:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/proc/irq:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/proc/pressure:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/proc/scsi:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/proc/sys/debug:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/proc/sys/dev:ro,nodev,noexec,nosuid",
-    "bind+/proc/sys/fs:/proc/sys/fs:ro,nodev,noexec,nosuid",
-    "bind+/proc/sys/kernel:/proc/sys/kernel:ro,nodev,noexec,nosuid",
-    "bind+/proc/sys/vm:/proc/sys/vm:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/proc/sysvipc:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/proc/tty:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/sys/dev/block:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/sys/devices/virtual/powercap:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/sys/firmware:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/sys/fs/bpf:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/sys/fs/ext4:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/sys/fs/fuse:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/sys/fs/pstore:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/sys/fs/selinux:ro,nodev,noexec,nosuid",
-    "bind+/var/empty:/sys/fs/smackfs:ro,nodev,noexec,nosuid",
 ];
 
 /// Landlock sandboxing profile, activated with:
@@ -630,17 +582,14 @@ pub const PROFILE_LANDLOCK: &[&str] = &[
     // /proc is always allowed.
     "allow/lock/read,readdir+/run",
     "allow/lock/read,readdir,exec+/sbin",
-    "allow/lock/read,readdir,exec+/snap",
     "allow/lock/read,readdir+/sys",
     "allow/lock/read,readdir,exec+/usr",
     "allow/lock/read,readdir+/var",
-    "allow/lock/read,write,truncate+/dev/fd",
     "allow/lock/read,write+/dev/full",
     // /dev/null is always allowed.
     "allow/lock/read+/dev/random",
     "allow/lock/read+/dev/urandom",
     "allow/lock/read,write+/dev/zero",
-    "allow/lock/read,readdir,write+/selinux/context",
     "allow/lock/read,readdir,exec,write,create,delete,rename,symlink,truncate,mkdir,rmdir,mkfifo+/dev/shm",
     "allow/lock/read,readdir,exec,write,create,delete,rename,symlink,truncate,mkdir,rmdir,mkfifo,bind+/tmp",
     "allow/lock/read,readdir,exec,write,create,delete,rename,symlink,truncate,mkdir,rmdir,mkfifo,bind+/var/tmp",
@@ -651,9 +600,10 @@ pub const PROFILE_LANDLOCK: &[&str] = &[
 pub const PROFILE_PALUDIS: &[&str] = &[
     "include_profile linux",
     "include_profile tty",
-    "name/host:localhost",   // only effective when combined with unshare/uts:1
-    "lock:exec",             // esandbox
-    "log/verbose:false",     // verbose logging is intended for malware analysis.
+    "default/lock:warn", // set Landlock compat-level to best-effort when stacked with Landlock profile.
+    "name/host:localhost", // only effective when combined with unshare/uts:1
+    "lock:exec",         // esandbox
+    "log/verbose:false", // verbose logging is intended for malware analysis.
     "trace/memory_access:0", // fallback to unsafe proc_pid_mem(5) if cross memory attach fails.
     "trace/allow_safe_bind:true",
     "trace/allow_safe_kcapi:true",
@@ -668,6 +618,7 @@ pub const PROFILE_PALUDIS: &[&str] = &[
     "trace/allow_unsafe_memory:true",
     "trace/allow_unsafe_msgsnd:true",
     "trace/allow_unsafe_nice:true",
+    "trace/allow_unsafe_nomseal:true", // mseal(2) is a mitigation against compromised Syd.
     "trace/allow_unsafe_nopie:true",
     "trace/allow_unsafe_open_path:true",
     "trace/allow_unsafe_perf:true",
@@ -694,11 +645,34 @@ pub const PROFILE_PALUDIS: &[&str] = &[
     "sandbox/tpe:off",
     // Filter noisy systemd access.
     "filter/net/connect+/run/systemd/userdb/io.systemd.DynamicUser",
+    // Defaults for /selinux.
+    "allow/stat,chdir,readdir+/selinux",
+    "allow/stat,chdir,readdir+/selinux/context",
+    "allow/read,stat,chdir,readdir,write+/selinux/context/**",
+    // Defaults for sysfs.
+    "allow/stat,chdir,readdir+/sys",
+    "allow/stat,chdir,readdir+/sys/devices",
+    "allow/stat,chdir,readdir+/sys/devices/system",
+    "allow/stat,chdir,readdir+/sys/devices/system/cpu",
+    "allow/read,stat+/sys/devices/system/cpu/isolated",
+    "allow/read,stat+/sys/devices/system/cpu/kernel_max",
+    "allow/read,stat+/sys/devices/system/cpu/online",
+    "allow/read,stat+/sys/devices/system/cpu/offline",
+    "allow/read,stat+/sys/devices/system/cpu/possible",
+    "allow/read,stat+/sys/devices/system/cpu/present",
+    "allow/stat,chdir,readdir+/sys/fs",
+    "allow/read,stat,chdir,readdir+/sys/fs/cgroup/***",
+    "allow/stat,chdir,readdir+/sys/kernel",
+    "allow/stat,chdir,readdir+/sys/kernel/mm",
+    "allow/read,stat,chdir,readdir+/sys/kernel/mm/transparent_hugepage/***",
+    "allow/stat,chdir,readdir+/sys/kernel/security",
+    "allow/read,stat,chdir,readdir+/sys/kernel/security/apparmor/***",
     // Defaults for procfs.
     "allow/read,stat+/proc/crypto",
     "allow/read,stat+/proc/cpuinfo",
     "allow/read,stat+/proc/loadavg",
     "allow/read,stat+/proc/meminfo",
+    "allow/read,stat+/proc/stat",
     "allow/read,stat+/proc/uptime",
     "allow/read,stat+/proc/version",
     "mask-/proc/version", // override mask in linux profile.
@@ -711,6 +685,19 @@ pub const PROFILE_PALUDIS: &[&str] = &[
     // by readdir'ing /proc hence we allow it here but deny
     // in the default secure `linux` profile.
     "allow/stat+/proc/[0-9]*",
+    "allow/read,stat+/proc/filesystems",
+    "allow/read,stat+/proc/self/mounts",
+    "allow/read,stat+/proc/self/task/[0-9]*/mounts",
+    "allow/read,stat+/proc/mounts", // symlink to self/mounts
+    "allow/read,stat,chdir,readdir+/proc/self/attr/***",
+    "allow/read,stat,chdir,readdir+/proc/self/task/[0-9]*/attr/***",
+    "allow/read,stat,chdir,readdir+/proc/sys",
+    "allow/read,stat,chdir,readdir+/proc/sys/fs/***",
+    "allow/read,stat,chdir,readdir+/proc/sys/kernel/***",
+    "allow/read,stat,chdir,readdir+/proc/sys/vm/***",
+    // Allow /var/cache for Paludis builds.
+    "allow/read,stat,chdir,readdir,mktemp+/var/cache/***",
+    "allow/write,create,delete,mkdir,rmdir,mkfifo,rename,symlink,truncate,chown,chgrp,chmod,chattr,utime+/var/cache/**",
     // Defaults for network sandboxing.
     "allow/net/bind+loopback!0",
     "allow/net/bind+loopback!1024-65535",
@@ -730,13 +717,8 @@ pub const PROFILE_PALUDIS: &[&str] = &[
 /// UID, GID, USER, and HOME before parsing this
 /// profile.
 pub const PROFILE_USER: &[&str] = &[
-    "include_profile landlock",
     "include_profile linux",
     "include_profile tty",
-    "include_profile immutable",
-    "unshare/all:try",  // use namespaces only if supported.
-    "bind-/home:/home", // immutable mounts this ro.
-    "bind+/home:/home:rw,nodev,nosuid",
     // dmesg(8)
     "trace/allow_safe_syslog:true",
     // TPE
@@ -747,9 +729,6 @@ pub const PROFILE_USER: &[&str] = &[
     // Disallow s{u,g}id bits on files.
     // Disallow setting group+other bits.
     "trace/force_umask:7077",
-    // /etc
-    "allow/read,stat,chdir,readdir+/etc/*sh*/***",
-    "allow/read,stat,chdir,readdir+/etc/profile*/***",
     // /home
     // 1. We disallow editing dotfiles.
     // 2. We mark shell-history files as append-only.
@@ -767,25 +746,17 @@ pub const PROFILE_USER: &[&str] = &[
     "allow/read,stat,chdir,readdir+/dev/shm",
     "allow/read,stat,chdir,readdir+/run",
     "allow/read,stat,chdir,readdir+/var/run",
-    "allow/lock/all+/dev/shm", // immutable mounts a tmpfs here.
-    "allow/lock/all+/run",     // ditto.
-    "allow/lock/all+/var/run", // ditto.
-    "allow/lock/all+${HOME}",
     "allow/read,stat,chdir,readdir+${HOME}/***",
     "allow/all+${HOME}/**",
     "allow/net/bind+${HOME}/**",
     "allow/net/connect+${HOME}/**",
     "allow/net/sendfd+${HOME}/**",
     // /run/user/uid
-    "allow/lock/all+/run/user/${SYD_UID}",
     "allow/all+/run/user/${SYD_UID}/**",
     "allow/read,stat,chdir,readdir+/run/user/${SYD_UID}/***",
     "allow/net/connect+/run/user/${SYD_UID}/**",
     // Allow access to DRI and sound.
     "allow/read,stat,chdir,readdir+/proc/asound/***",
-    "allow/lock/read,readdir,write,ioctl+/dev/dri",
-    "allow/lock/read,readdir,write,ioctl+/dev/input",
-    "allow/lock/read,readdir,write,ioctl+/dev/snd",
     "allow/read,stat,chdir,readdir,write,ioctl+/dev/dri/**",
     "allow/read,stat,chdir,readdir,write,ioctl+/dev/input/**",
     "allow/read,stat,chdir,readdir,write,ioctl+/dev/snd/**",
@@ -835,14 +806,16 @@ pub const PROFILE_LINUX: &[&str] = &[
     // ie identifying fellow processing by readdir'ing `/proc`.
     // Notably, this is allowed in the `paludis` profile.
     "allow/read,stat,chdir,readdir+/proc",
-    "allow/read,stat+/proc/filesystems",
-    // SAFETY: We allow basic stat access (global, per-process, per-task).
+    // SAFETY: We allow basic stat access (not global but per-process and per-task).
     // trace/allow_unsafe_magiclinks:0 default is another layer of
     // protection against potential malicious activity with proc
-    // magiclinks.
-    "allow/read,stat+/proc/stat",
+    // magiclinks. Paludis profile allows global stat access for /proc/stat.
     "allow/read,stat+/proc/self/stat",
     "allow/read,stat+/proc/self/task/[0-9]*/stat",
+    "allow/read,stat+/proc/self/statm",
+    "allow/read,stat+/proc/self/task/[0-9]*/statm",
+    "allow/read,stat+/proc/self/status",
+    "allow/read,stat+/proc/self/task/[0-9]*/status",
     // SAFETY: We provide an empty file in place of /proc/cmdline
     // for compatibility rather than denying outright. In any case,
     // syd does not leak the contents of this file to the sandbox
@@ -853,15 +826,10 @@ pub const PROFILE_LINUX: &[&str] = &[
     // This is consistent with uname(2) restrictions.
     "allow/read,stat+/proc/version",
     "mask+/proc/version",
-    "allow/read,stat,chdir,readdir+/proc/sys",
-    "allow/read,stat,chdir,readdir+/proc/sys/fs/***",
-    "allow/read,stat,chdir,readdir+/proc/sys/kernel/***",
-    "allow/read,stat,chdir,readdir+/proc/sys/vm/***",
     // SAFETY: Modification of /proc/sys/kernel/modprobe requires no
     // capabilities (can cause arbitrary code to be inserted into the
     // kernel via a replacement modprobe)!
     // https://forums.grsecurity.net/viewtopic.php?f=7&t=2522
-    "deny/read,stat+/proc/sys/kernel/modprobe",
     "mask+/proc/sys/kernel/modprobe",
     // SAFETY: Provide an empty file in place of /sys/kernel/notes.
     // See: https://lwn.net/Articles/962782/
@@ -882,56 +850,23 @@ pub const PROFILE_LINUX: &[&str] = &[
     "allow/stat,chdir,readdir+/proc/self/task/[0-9]*",
     "allow/stat,chdir,readdir+/proc/self/task/[0-9]*/fd",
     "allow/stat,chdir,readdir+/proc/self/task/[0-9]*/fdinfo",
-    "allow/read,stat,chdir,readdir,write,truncate+/proc/self/fd/[0-9]*",
-    "allow/read,stat,chdir,readdir,write,truncate+/proc/self/task/[0-9]*/fd/[0-9]*",
-    r"allow/read,stat,write,ioctl+/proc/self/fd/pipe:\[[0-9]*\]",
-    r"allow/read,stat,write,ioctl+/proc/self/fd/socket:\[[0-9]*\]",
-    r"allow/read,stat,write,ioctl+/proc/self/task/[0-9]*/fd/pipe:\[[0-9]*\]",
-    r"allow/read,stat,write,ioctl+/proc/self/task/[0-9]*/fd/socket:\[[0-9]*\]",
+    "allow/read,stat,write,truncate,ioctl+/proc/self/fd/[0-9]*",
+    "allow/read,stat,write,truncate,ioctl+/proc/self/task/[0-9]*/fd/[0-9]*",
+    r"allow/read,stat,write,truncate,ioctl+/proc/self/fd/pipe:\[[0-9]*\]",
+    r"allow/read,stat,write,truncate,ioctl+/proc/self/fd/socket:\[[0-9]*\]",
+    r"allow/read,stat,write,truncate,ioctl+/proc/self/task/[0-9]*/fd/pipe:\[[0-9]*\]",
+    r"allow/read,stat,write,truncate,ioctl+/proc/self/task/[0-9]*/fd/socket:\[[0-9]*\]",
     "allow/read,stat,chdir,readdir+/proc/self/fdinfo/[0-9]*",
     "allow/read,stat,chdir,readdir+/proc/self/task/[0-9]*/fdinfo/[0-9]*",
-    "allow/read,stat,chdir,readdir+/proc/self/attr/***",
-    "allow/read,stat,chdir,readdir+/proc/self/task/[0-9]*/attr/***",
     "allow/read,stat+/proc/self/maps",
     "allow/read,stat+/proc/self/task/[0-9]*/maps",
-    "allow/read,stat+/proc/self/mounts",
-    "allow/read,stat+/proc/self/task/[0-9]*/mounts",
-    "allow/read,stat+/proc/mounts", // symlink to self/mounts
     "allow/stat,chdir,readdir+/proc/self/root",
     "allow/stat,chdir,readdir+/proc/self/task/[0-9]*/root",
-    "allow/read,stat+/proc/self/stat",
-    "allow/read,stat+/proc/self/task/[0-9]*/stat",
-    "allow/read,stat+/proc/self/statm",
-    "allow/read,stat+/proc/self/task/[0-9]*/statm",
-    "allow/read,stat+/proc/self/status",
-    "allow/read,stat+/proc/self/task/[0-9]*/status",
-    // /sys
-    "allow/stat,chdir,readdir+/sys",
-    "allow/stat,chdir,readdir+/sys/devices",
-    "allow/stat,chdir,readdir+/sys/devices/system",
-    "allow/stat,chdir,readdir+/sys/devices/system/cpu",
-    "allow/read,stat+/sys/devices/system/cpu/isolated",
-    "allow/read,stat+/sys/devices/system/cpu/kernel_max",
-    "allow/read,stat+/sys/devices/system/cpu/online",
-    "allow/read,stat+/sys/devices/system/cpu/offline",
-    "allow/read,stat+/sys/devices/system/cpu/possible",
-    "allow/read,stat+/sys/devices/system/cpu/present",
-    "allow/stat,chdir,readdir+/sys/fs",
-    "allow/read,stat,chdir,readdir+/sys/fs/cgroup/***",
-    "allow/stat,chdir,readdir+/sys/kernel",
-    "allow/stat,chdir,readdir+/sys/kernel/mm",
-    "allow/read,stat,chdir,readdir+/sys/kernel/mm/transparent_hugepage/***",
-    "allow/stat,chdir,readdir+/sys/kernel/security",
-    "allow/read,stat,chdir,readdir+/sys/kernel/security/apparmor/***",
     // /run
     "allow/stat,chdir,readdir+/run",
     "allow/stat,chdir,readdir+/run/systemd",
     "allow/stat,chdir,readdir+/run/systemd/resolve",
     "allow/read,stat+/run/systemd/resolve/*.conf",
-    // /selinux
-    "allow/stat,chdir,readdir+/selinux",
-    "allow/stat,chdir,readdir+/selinux/context",
-    "allow/read,stat,chdir,readdir,write+/selinux/context/**",
     // /tmp and /var/tmp
     "allow/read,stat,chdir,readdir,mktemp+/tmp/***",
     "allow/read,stat,chdir,readdir,mktemp+/var/tmp/***",
@@ -940,8 +875,6 @@ pub const PROFILE_LINUX: &[&str] = &[
     // /var
     "allow/stat,chdir,readdir+/var",
     "allow/stat,chdir,readdir,chroot+/var/empty",
-    "allow/read,stat,chdir,readdir,mktemp+/var/cache/***",
-    "allow/write,create,delete,mkdir,rmdir,mkfifo,rename,symlink,truncate,chown,chgrp,chmod,chattr,utime+/var/cache/**",
     "allow/read,stat+/var/lib/sss/mc/passwd", // Required for LDAP.
     "allow/read,stat+/var/lib/sss/mc/group",  // ditto.
     // /etc
@@ -952,6 +885,7 @@ pub const PROFILE_LINUX: &[&str] = &[
     "allow/read,stat,chdir,readdir+/etc/ca-certificates/***",
     "allow/read,stat,chdir,readdir+/etc/env.d/***",
     "allow/read,stat,chdir,readdir+/etc/groff/***",
+    "allow/read,stat,chdir,readdir+/etc/ld.so.conf.d/***",
     "allow/read,stat+/etc/environment",
     "allow/read,stat+/etc/ethertypes",
     "allow/read,stat+/etc/group",
@@ -983,7 +917,7 @@ pub const PROFILE_LINUX: &[&str] = &[
     // /home
     //
     // Do _not_ allow readdir which allows enumerating other users!
-    "allow/read,stat,chdir+/home",
+    "allow/stat,chdir+/home",
     // Defaults for Network sandboxing:
     //
     // Allow network access to unnamed UNIX sockets.
@@ -1207,6 +1141,8 @@ pub const GET_ID_SYSCALLS: &[&str] = &[
     "getresuid32",
     "getresgid",
     "getresgid32",
+    "getgroups",
+    "getgroups32",
 ];
 
 /// The list of system calls which are of the futex family.
@@ -2122,6 +2058,7 @@ pub const AES_SYSCALLS: &[&str] = &[
 pub const AES_FCNTL_OPS: &[u64] = &[
     libc::F_GETFD as u64,
     libc::F_SETFD as u64,
+    libc::F_ADD_SEALS as u64,
     libc::F_OFD_SETLKW as u64,
 ];
 
@@ -2237,10 +2174,16 @@ pub(crate) static HAVE_STATX_MNT_ID_UNIQUE: Lazy<bool> = Lazy::new(|| {
     major > 6 || (major == 6 && minor >= 8)
 });
 
+/// Check for AT_EXECVE_CHECK support (Linux-6.14 or newer).
+pub static HAVE_AT_EXECVE_CHECK: Lazy<bool> = Lazy::new(|| {
+    let (major, minor) = *KERNEL_VERSION;
+    major > 6 || (major == 6 && minor >= 14)
+});
+
 //
 // Below is internal territory, you have been warned.
-//
-/// Determine minimum allowed mmap address by reading `/proc/sys/vm/mmap_min_addr`.
+
+// Determine minimum allowed mmap address by reading `/proc/sys/vm/mmap_min_addr`.
 pub(crate) static MMAP_MIN_ADDR: Lazy<u64> = Lazy::new(|| proc_mmap_min_addr().unwrap_or(4096));
 
 // Trace data for ptrace(2) hooked system calls.
@@ -2311,37 +2254,6 @@ pub(crate) fn ROOT_FILE() -> BorrowedFd<'static> {
     unsafe { BorrowedFd::borrow_raw(ROOT_FD()) }
 }
 
-/// Returns a reference to the static `/dev` dirfd.
-///
-/// Calling this before calling `proc_init` will panic!
-#[allow(clippy::disallowed_methods)]
-#[allow(non_snake_case)]
-#[inline(always)]
-pub(crate) fn DEV_FD() -> RawFd {
-    *_DEV_FD.get().unwrap()
-}
-
-/// Returns a reference to the static `/dev` unique mount id.
-///
-/// Calling this before calling `proc_init` will panic!
-#[allow(clippy::disallowed_methods)]
-#[allow(non_snake_case)]
-#[inline(always)]
-pub(crate) fn DEV_MNT_ID() -> u64 {
-    *_DEV_MNT_ID.get().unwrap()
-}
-
-/// Returns a reference to the static `/dev` dirfd.
-///
-/// Calling this before calling `proc_init` will panic!
-#[allow(clippy::disallowed_methods)]
-#[allow(non_snake_case)]
-#[inline(always)]
-pub(crate) fn DEV_FILE() -> BorrowedFd<'static> {
-    // SAFETY: `proc_init' is called beforehand.
-    unsafe { BorrowedFd::borrow_raw(DEV_FD()) }
-}
-
 /// Returns a reference to the static `/proc` dirfd.
 ///
 /// Calling this before calling `proc_init` will panic!
@@ -2373,37 +2285,6 @@ pub(crate) fn PROC_FILE() -> BorrowedFd<'static> {
     unsafe { BorrowedFd::borrow_raw(PROC_FD()) }
 }
 
-/// Returns a reference to the static `/sys` dirfd.
-///
-/// Calling this before calling `proc_init` will panic!
-#[allow(clippy::disallowed_methods)]
-#[allow(non_snake_case)]
-#[inline(always)]
-pub fn SYS_FD() -> RawFd {
-    *_SYS_FD.get().unwrap()
-}
-
-/// Returns a reference to the static `/sys` unique mount id.
-///
-/// Calling this before calling `proc_init` will panic!
-#[allow(clippy::disallowed_methods)]
-#[allow(non_snake_case)]
-#[inline(always)]
-pub(crate) fn SYS_MNT_ID() -> u64 {
-    *_SYS_MNT_ID.get().unwrap()
-}
-
-/// Returns a reference to the static `/sys` dirfd.
-///
-/// Calling this before calling `proc_init` will panic!
-#[allow(clippy::disallowed_methods)]
-#[allow(non_snake_case)]
-#[inline(always)]
-pub(crate) fn SYS_FILE() -> BorrowedFd<'static> {
-    // SAFETY: `proc_init' is called beforehand.
-    unsafe { BorrowedFd::borrow_raw(SYS_FD()) }
-}
-
 /// Returns a reference to the static `/dev/null` fd.
 ///
 /// Calling this before calling `proc_init` will panic!
@@ -2443,24 +2324,12 @@ pub(crate) static _ROOT_FD: OnceLock<RawFd> = OnceLock::new();
 /// Unique mount id to `/`, ie the root file system.
 pub(crate) static _ROOT_MNT_ID: OnceLock<u64> = OnceLock::new();
 
-/// File descriptor to /dev file system.
-pub(crate) static _DEV_FD: OnceLock<RawFd> = OnceLock::new();
-
-/// Unique mount id to /dev file system.
-pub(crate) static _DEV_MNT_ID: OnceLock<u64> = OnceLock::new();
-
 /// File descriptor to /proc file system.
 pub(crate) static _PROC_FD: OnceLock<RawFd> = OnceLock::new();
 
 /// Unique mount id to /proc file system.
 pub(crate) static _PROC_MNT_ID: OnceLock<u64> = OnceLock::new();
 
-/// File descriptor to /sys file system.
-pub(crate) static _SYS_FD: OnceLock<RawFd> = OnceLock::new();
-
-/// Unique mount id to /sys file system.
-pub(crate) static _SYS_MNT_ID: OnceLock<u64> = OnceLock::new();
-
 /// File descriptor to /dev/null character device.
 pub(crate) static _NULL_FD: OnceLock<RawFd> = OnceLock::new();
 
@@ -2477,14 +2346,16 @@ pub fn proc_init() -> Result<(), Errno> {
         STATX_MNT_ID
     };
 
-    let fd_root = nix::fcntl::open("/", OFlag::O_RDONLY | OFlag::O_DIRECTORY, Mode::empty())?;
+    let fd_root = nix::fcntl::open(
+        "/",
+        OFlag::O_PATH | OFlag::O_DIRECTORY | OFlag::O_CLOEXEC,
+        Mode::empty(),
+    )?;
     let id_root = fstatx(&fd_root, mask).map(|stx| stx.stx_mnt_id)?;
 
     // SAFETY: To make this file descriptor harder to spot by an
     // attacker we duplicate it to a random fd number.
-    let fd_rand = duprand(fd_root)?;
-    let _ = close(fd_root);
-    let fd_root = fd_rand;
+    let fd_root = duprand(fd_root.as_raw_fd(), OFlag::O_CLOEXEC)?;
 
     info!("ctx": "run", "op": "opendir_root",
         "msg": "opened root directory",
@@ -2493,40 +2364,20 @@ pub fn proc_init() -> Result<(), Errno> {
     _ROOT_FD.set(fd_root).or(Err(Errno::EAGAIN))?;
     _ROOT_MNT_ID.set(id_root).or(Err(Errno::EAGAIN))?;
 
-    let fd_dev = nix::fcntl::openat(
-        Some(fd_root),
-        XPath::from_bytes(b"dev"),
-        OFlag::O_DIRECTORY,
-        Mode::empty(),
-    )?;
-    let id_dev = fstatx(&fd_dev, mask).map(|stx| stx.stx_mnt_id)?;
-
-    // SAFETY: To make this file descriptor harder to spot by an
-    // attacker we duplicate it to a random fd number.
-    let fd_rand = duprand(fd_dev)?;
-    let _ = close(fd_dev);
-    let fd_dev = fd_rand;
-
-    info!("ctx": "run", "op": "opendir_dev",
-        "msg": "opened /dev directory",
-        "fd": fd_dev,
-        "id": id_dev);
-    _DEV_FD.set(fd_dev).or(Err(Errno::EAGAIN))?;
-    _DEV_MNT_ID.set(id_dev).or(Err(Errno::EAGAIN))?;
+    // SAFETY: fd_root is a valid fd for process lifetime.
+    let fd_root = unsafe { BorrowedFd::borrow_raw(fd_root) };
 
     let fd_proc = nix::fcntl::openat(
-        Some(fd_root),
+        fd_root,
         XPath::from_bytes(b"proc"),
-        OFlag::O_DIRECTORY,
+        OFlag::O_PATH | OFlag::O_DIRECTORY | OFlag::O_CLOEXEC,
         Mode::empty(),
     )?;
     let id_proc = fstatx(&fd_proc, mask).map(|stx| stx.stx_mnt_id)?;
 
     // SAFETY: To make this file descriptor harder to spot by an
     // attacker we duplicate it to a random fd number.
-    let fd_rand = duprand(fd_proc)?;
-    let _ = close(fd_proc);
-    let fd_proc = fd_rand;
+    let fd_proc = duprand(fd_proc.as_raw_fd(), OFlag::O_CLOEXEC)?;
 
     info!("ctx": "run", "op": "opendir_proc",
         "msg": "opened /proc directory",
@@ -2535,43 +2386,20 @@ pub fn proc_init() -> Result<(), Errno> {
     _PROC_FD.set(fd_proc).or(Err(Errno::EAGAIN))?;
     _PROC_MNT_ID.set(id_proc).or(Err(Errno::EAGAIN))?;
 
-    let fd_sys = nix::fcntl::openat(
-        Some(fd_root),
-        XPath::from_bytes(b"sys"),
-        OFlag::O_DIRECTORY,
-        Mode::empty(),
-    )?;
-    let id_sys = fstatx(&fd_sys, mask).map(|stx| stx.stx_mnt_id)?;
-
-    // SAFETY: To make this file descriptor harder to spot by an
-    // attacker we duplicate it to a random fd number.
-    let fd_rand = duprand(fd_sys)?;
-    let _ = close(fd_sys);
-    let fd_sys = fd_rand;
-
-    info!("ctx": "run", "op": "opendir_sys",
-        "msg": "opened /sys directory",
-        "fd": fd_sys,
-        "id": id_sys);
-    _SYS_FD.set(fd_sys).or(Err(Errno::EAGAIN))?;
-    _SYS_MNT_ID.set(id_sys).or(Err(Errno::EAGAIN))?;
-
     let fd_null = nix::fcntl::openat(
-        Some(fd_dev),
-        XPath::from_bytes(b"null"),
-        OFlag::empty(),
+        fd_root,
+        XPath::from_bytes(b"dev/null"),
+        OFlag::O_PATH | OFlag::O_CLOEXEC,
         Mode::empty(),
     )?;
     let id_null = fstatx(&fd_null, mask).map(|stx| stx.stx_mnt_id)?;
 
     // SAFETY: Validate what we've opened is indeed `/dev/null`.
-    crate::validate_dev_null(fd_null)?;
+    crate::validate_dev_null(fd_null.as_raw_fd())?;
 
     // SAFETY: To make this file descriptor harder to spot by an
     // attacker we duplicate it to a random fd number.
-    let fd_rand = duprand(fd_null)?;
-    let _ = close(fd_null);
-    let fd_null = fd_rand;
+    let fd_null = duprand(fd_null.as_raw_fd(), OFlag::O_CLOEXEC)?;
 
     info!("ctx": "run", "op": "opendev_null",
         "msg": "opened /dev/null",
@@ -2590,22 +2418,49 @@ pub fn proc_close() {
         let _ = close(*fd);
     }
 
-    if let Some(fd) = _DEV_FD.get() {
-        let _ = close(*fd);
-    }
-
     if let Some(fd) = _PROC_FD.get() {
         let _ = close(*fd);
     }
 
-    if let Some(fd) = _SYS_FD.get() {
-        let _ = close(*fd);
-    }
-
     if let Some(fd) = _NULL_FD.get() {
         let _ = close(*fd);
     }
 }
 
+// FileMap populated with static FDs.
+// safe_canonicalize clones this on each call.
+pub(crate) static FILEMAP: Lazy<FileMap> = Lazy::new(|| {
+    let mut filemap: FileMap = FileMap::with_capacity(FILECAP);
+
+    let entry = FileMapEntry::new(
+        ROOT_FD().into(),
+        Some(FileType::Dir),
+        Some(ROOT_MNT_ID()),
+        None,
+    );
+    filemap.0.insert(XPathBuf::from("/"), entry);
+
+    let entry = FileMapEntry::new(
+        PROC_FD().into(),
+        Some(FileType::Dir),
+        Some(PROC_MNT_ID()),
+        None,
+    );
+    filemap.0.insert(XPathBuf::from("/proc"), entry);
+
+    let entry = FileMapEntry::new(
+        NULL_FD().into(),
+        Some(FileType::Chr),
+        Some(NULL_MNT_ID()),
+        None,
+    );
+    filemap.0.insert(XPathBuf::from("/dev/null"), entry);
+
+    filemap
+});
+
+// FileMap default capacity.
+pub(crate) const FILECAP: usize = 8;
+
 /// Reference to the global PidFd map.
 pub static PIDFD_MAP: OnceLock<PidFdMap> = OnceLock::new();
