@@ -22,10 +22,10 @@
 //! [![syd](https://git.sr.ht/~alip/syd/blob/main/data/syd.png)](https://todo.sr.ht/~alip/syd)
 //! [![GNU](https://web.archive.org/web/20221222061733if_/https://dev.exherbo.org/~alip/images/gnu.png)](https://www.gnu.org/philosophy/philosophy.html)
 //! [![Linux](https://chesswob.org/jja/tux.png)](https://www.kernel.org/category/about.html)
-//! [![Exherbo](https://web.archive.org/web/20230518155203if_/https://dev.exherbo.org/~alip/images/zebrapig.png)](https://www.exherbolinux.org/docs/gettingstarted.html)
+//! [![Exherbo](https://web.archive.org/web/20230518155203if_/https://dev.exherbo.org/~alip/images/zebrapig.png)](https://www.exherbo.org/docs/gettingstarted.html)
 //! [![musl libc](https://www.chesswob.org/jja/musl-inside.png)](https://www.musl-libc.org/)
 //! [![libsecc☮mp](https://web.archive.org/web/20221222061720if_/https://dev.exherbo.org/~alip/images/libseccomp.png)](https://github.com/seccomp/libseccomp)
-//! [![Paludis](http://paludis.exherbolinux.org/paludis_270.png)](https://paludis.exherbolinux.org)
+//! [![Paludis](http://paludis.exherbo.org/paludis_270.png)](https://paludis.exherbo.org)
 //!
 //! syd is a **seccomp**(2) based sandboxing utility for modern Linux\[\>=5.6\]
 //! machines to sandbox unwanted process access to filesystem and network resources.
@@ -109,6 +109,10 @@
 // TODO: remove the allow-lint below when libseccomp is back to upstream.
 #![allow(macro_expanded_macro_exports_accessed_by_absolute_paths)]
 
+/// System call handlers
+#[macro_use]
+pub(crate) mod kernel;
+
 /// Utilities for caching
 pub mod cache;
 /// Compatibility code for different libcs
@@ -128,10 +132,12 @@ pub mod fs;
 pub mod hash;
 /// Secure computing hooks
 pub mod hook;
+/// Landlock policy helper library for Syd
+pub mod landlock_policy;
 /// Simple logging on standard error using JSON lines
 pub mod log;
-/// /proc nom parsers
-pub(crate) mod parsers;
+/// /proc and syd(2) nom parsers
+pub mod parsers;
 /// Path handling for UNIX
 pub mod path;
 /// /proc utilities
@@ -142,6 +148,8 @@ pub mod ptrace;
 pub mod sandbox;
 /// Execute program as sealed anonymous file
 pub mod seal;
+/// SealBox<T> for type-safe sealing/protecting
+pub mod sealbox;
 /// Interface to Linux prctl(2) speculation misfeature interfac
 pub mod spec;
 /// sysinfo(2) interface
@@ -186,7 +194,6 @@ extern crate lazy_static;
 #[allow(clippy::type_complexity)]
 #[allow(clippy::undocumented_unsafe_blocks)]
 pub mod landlock;
-/// Read the ELF dependency tree
 #[allow(dead_code)]
 #[allow(clippy::disallowed_methods)]
 #[allow(clippy::manual_flatten)]
@@ -199,29 +206,24 @@ pub mod unshare;
 
 use std::{
     arch::asm,
-    collections::{HashMap, HashSet},
-    ffi::{CStr, OsStr},
-    ops::{
-        BitAnd, BitAndAssign, BitOr, BitOrAssign, BitXor, BitXorAssign, Not, RangeInclusive, Sub,
-        SubAssign,
-    },
-    os::fd::{AsFd, AsRawFd, BorrowedFd, FromRawFd, OwnedFd, RawFd},
+    ffi::OsStr,
+    ops::{BitAnd, BitAndAssign, BitOr, BitOrAssign, BitXor, BitXorAssign, Not, Sub, SubAssign},
+    os::fd::{AsFd, AsRawFd, BorrowedFd, RawFd},
     path::Path,
     process::exit,
     str::FromStr,
 };
 
-use ahash::RandomState;
 use landlock::{
     path_beneath_rules, Access, AccessFs, AccessNet, NetPort, RestrictionStatus, Ruleset,
-    RulesetAttr, RulesetCreatedAttr, RulesetError, ScopeFlag,
+    RulesetAttr, RulesetCreatedAttr, RulesetError, Scope,
 };
 use lexis::ToName;
 use libseccomp::{ScmpAction, ScmpArch, ScmpFilterContext, ScmpSyscall, ScmpVersion};
 use nix::{
     dir::Dir,
     errno::Errno,
-    fcntl::{open, OFlag},
+    fcntl::OFlag,
     libc::c_int,
     mount::MsFlags,
     sched::{unshare, CloneFlags},
@@ -241,8 +243,10 @@ use crate::{
     compat::{fstatx, lsm_list_modules, waitid, STATX_BASIC_STATS},
     err::{err2no, SydResult},
     fs::safe_clone,
-    landlock::{CreateRulesetError, PathBeneath, PathFd, RulesetStatus, ABI},
+    hash::{aes_ctr_info, hmac_sha256_info},
+    landlock::{RulesetStatus, ABI},
     path::{mask_path, XPath, XPathBuf},
+    sealbox::check_mseal_support,
     spec::{speculation_get, SpeculationFeature},
 };
 
@@ -343,13 +347,13 @@ impl std::fmt::Display for SydMemoryMap {
             MMapPath::Path(path) => mask_path(path),
             MMapPath::Heap => "[heap]".to_string(),
             MMapPath::Stack => "[stack]".to_string(),
-            MMapPath::TStack(tid) => format!("[stack:{}]", tid),
+            MMapPath::TStack(tid) => format!("[stack:{tid}]"),
             MMapPath::Vdso => "[vdso]".to_string(),
             MMapPath::Vvar => "[vvar]".to_string(),
             MMapPath::Vsyscall => "[vsyscall]".to_string(),
             MMapPath::Rollup => "[rollup]".to_string(),
             MMapPath::Anonymous => "[anon]".to_string(),
-            MMapPath::Vsys(key) => format!("[vsys:{}]", key),
+            MMapPath::Vsys(key) => format!("[vsys:{key}]"),
             MMapPath::Other(pseudo_path) => mask_path(Path::new(pseudo_path)),
         };
 
@@ -721,6 +725,16 @@ pub fn syd_info() {
         },
     );
 
+    println!(
+        "Memory sealing is {}supported.",
+        if check_mseal_support() { "" } else { "not " }
+    );
+
+    // Print ctr(aes) kernel support.
+    println!("{}", aes_ctr_info());
+    // Print hmac(sha256) kernel support.
+    println!("{}", hmac_sha256_info());
+
     let lsms = match lsm_list_modules() {
         Ok(lsms) => lsms
             .into_iter()
@@ -875,8 +889,8 @@ pub fn lock_enabled(abi: ABI) -> u8 {
 
         // Landlock network is ABI>=6.
         if abi as i32 >= ABI::V6 as i32 {
-            ruleset_ref.handle_access(ScopeFlag::AbstractUnixSocket)?;
-            ruleset_ref.handle_access(ScopeFlag::Signal)?;
+            ruleset_ref.scope(Scope::AbstractUnixSocket)?;
+            ruleset_ref.scope(Scope::Signal)?;
         }
 
         ruleset
@@ -1235,19 +1249,10 @@ impl Serialize for SydSigSet {
 
         for sig_num in 1..max_signal {
             if self.contains(sig_num) {
-                // SAFETY: strsignal returns a descriptive string for a
-                // valid signal.  It should never return NULL for a
-                // valid signal, but if it does, we fallback to the
-                // number.
-                let c_ptr = unsafe { libc::strsignal(sig_num) };
-                if !c_ptr.is_null() {
-                    // SAFETY: strsignal returned success,
-                    // we have a valid c string.
-                    let c_str = unsafe { CStr::from_ptr(c_ptr) };
-                    signals.push(c_str.to_string_lossy().into_owned());
-                } else {
-                    signals.push(sig_num.to_string());
-                }
+                let sig_name = Signal::try_from(sig_num)
+                    .map(|sig| sig.to_string())
+                    .unwrap_or_else(|_| format!("SIG{sig_num}"));
+                signals.push(sig_name);
             }
         }
 
@@ -1621,7 +1626,7 @@ pub const fn scmp_arch_raw(arch: ScmpArch) -> u32 {
         ScmpArch::Riscv64 => libseccomp_sys::SCMP_ARCH_RISCV64,
         ScmpArch::Sheb => libseccomp_sys::SCMP_ARCH_SHEB,
         ScmpArch::Sh => libseccomp_sys::SCMP_ARCH_SH,
-        _ => todo!(),
+        _ => unreachable!(),
     }
 }
 
@@ -1680,345 +1685,6 @@ pub fn nsflag_name(flag: libc::c_int) -> String {
     .to_string()
 }
 
-/// Data structure to store the landlock security policy.
-#[derive(Clone, Debug, Default)]
-pub struct LandlockPolicy {
-    /// Landlock read pathset
-    pub read_pathset: Option<HashSet<XPathBuf, RandomState>>,
-    /// Landlock write pathset
-    pub write_pathset: Option<HashSet<XPathBuf, RandomState>>,
-    /// Landlock execute pathset
-    pub exec_pathset: Option<HashSet<XPathBuf, RandomState>>,
-    /// Landlock ioctl(2) pathset
-    pub ioctl_pathset: Option<HashSet<XPathBuf, RandomState>>,
-    /// Landlock create pathset
-    pub create_pathset: Option<HashSet<XPathBuf, RandomState>>,
-    /// Landlock delete pathset
-    pub delete_pathset: Option<HashSet<XPathBuf, RandomState>>,
-    /// Landlock rename pathset
-    pub rename_pathset: Option<HashSet<XPathBuf, RandomState>>,
-    /// Landlock symlink pathset
-    pub symlink_pathset: Option<HashSet<XPathBuf, RandomState>>,
-    /// Landlock truncate pathset
-    pub truncate_pathset: Option<HashSet<XPathBuf, RandomState>>,
-    /// Landlock readdir pathset
-    pub readdir_pathset: Option<HashSet<XPathBuf, RandomState>>,
-    /// Landlock mkdir pathset
-    pub mkdir_pathset: Option<HashSet<XPathBuf, RandomState>>,
-    /// Landlock rmdir pathset
-    pub rmdir_pathset: Option<HashSet<XPathBuf, RandomState>>,
-    /// Landlock mkdev pathset
-    pub mkdev_pathset: Option<HashSet<XPathBuf, RandomState>>,
-    /// Landlock mkfifo pathset
-    pub mkfifo_pathset: Option<HashSet<XPathBuf, RandomState>>,
-    /// Landlock make socket pathset
-    pub bind_pathset: Option<HashSet<XPathBuf, RandomState>>,
-    /// Landlock bind portset
-    pub bind_portset: Option<HashSet<RangeInclusive<u16>, RandomState>>,
-    /// Landlock connect portset
-    pub conn_portset: Option<HashSet<RangeInclusive<u16>, RandomState>>,
-    /// Scoped abstract UNIX sockets
-    pub scoped_abs: bool,
-    /// Scoped UNIX signals
-    pub scoped_sig: bool,
-}
-
-// syd::landlock::path_beneath_rules tailored for Syd use-case.
-fn landlock_path_beneath_rules<I, P>(
-    paths: I,
-    access: AccessFs,
-) -> impl Iterator<Item = Result<PathBeneath<PathFd>, RulesetError>>
-where
-    I: IntoIterator<Item = P>,
-    P: AsRef<XPath>,
-{
-    paths.into_iter().filter_map(move |p| {
-        let p = p.as_ref();
-        match open(p, OFlag::O_PATH | OFlag::O_CLOEXEC, Mode::empty()) {
-            Ok(fd) => {
-                // SAFETY: open returns a valid FD.
-                Some(Ok(PathBeneath::new(
-                    PathFd {
-                        fd: unsafe { OwnedFd::from_raw_fd(fd) },
-                    },
-                    access,
-                )))
-            }
-            Err(errno @ Errno::ENOENT) => {
-                info!("ctx": "init", "op": "landlock_create_ruleset",
-                    "path": p, "access": format!("{access:?}"), "err": errno as i32,
-                    "msg": format!("open path `{p}' for Landlock failed: {errno}"));
-                None
-            }
-            Err(errno) => {
-                error!("ctx": "init", "op": "landlock_create_ruleset",
-                    "path": p, "access": format!("{access:?}"), "err": errno as i32,
-                    "msg": format!("open path `{p}' for Landlock failed: {errno}"));
-                Some(Err(RulesetError::CreateRuleset(
-                    CreateRulesetError::CreateRulesetCall {
-                        source: errno.into(),
-                    },
-                )))
-            }
-        }
-    })
-}
-
-/// A helper function to wrap the operations and reduce duplication.
-#[allow(clippy::arithmetic_side_effects)]
-#[allow(clippy::disallowed_methods)]
-pub fn landlock_operation(
-    abi: ABI,
-    policy: &LandlockPolicy,
-) -> Result<RestrictionStatus, RulesetError> {
-    // from_all includes IoctlDev of ABI >= 5 as necessary.
-    let mut ruleset = Ruleset::default().handle_access(AccessFs::from_all(abi))?;
-    let ruleset_ref = &mut ruleset;
-
-    // Network is ABI >= 4.
-    let mut network_rules_bind: HashSet<u16, RandomState> = HashSet::default();
-    if let Some(ref port_set) = policy.bind_portset {
-        for port_range in port_set {
-            for port in port_range.clone() {
-                network_rules_bind.insert(port);
-            }
-        }
-    }
-    if network_rules_bind.len() <= usize::from(u16::MAX) + 1 {
-        ruleset_ref.handle_access(AccessNet::BindTcp)?;
-    } else {
-        // SAFETY: All ports are allowed, do not handle the access right,
-        // rather than allowing each and every port.
-        network_rules_bind.clear();
-    }
-
-    let mut network_rules_conn: HashSet<u16, RandomState> = HashSet::default();
-    if let Some(ref port_set) = policy.conn_portset {
-        for port_range in port_set {
-            for port in port_range.clone() {
-                network_rules_conn.insert(port);
-            }
-        }
-    }
-    if network_rules_conn.len() <= usize::from(u16::MAX) + 1 {
-        ruleset_ref.handle_access(AccessNet::ConnectTcp)?;
-    } else {
-        // SAFETY: All ports are allowed, do not handle the access right,
-        // rather than allowing each and every port.
-        network_rules_conn.clear();
-    }
-
-    // Scopes are ABI >= 6.
-    if policy.scoped_abs {
-        ruleset_ref.handle_access(ScopeFlag::AbstractUnixSocket)?;
-    }
-    if policy.scoped_sig {
-        ruleset_ref.handle_access(ScopeFlag::Signal)?;
-    }
-
-    // Merge path rules based on access rights.
-    //
-    // Step 1: Accumulate all paths in a single set.
-    let mut all_pathset: HashSet<XPathBuf, RandomState> = HashSet::default();
-    if let Some(ref pathset) = policy.read_pathset {
-        all_pathset.extend(pathset.iter().cloned());
-    }
-    if let Some(ref pathset) = policy.write_pathset {
-        all_pathset.extend(pathset.iter().cloned());
-    }
-    if let Some(ref pathset) = policy.exec_pathset {
-        all_pathset.extend(pathset.iter().cloned());
-    }
-    if let Some(ref pathset) = policy.ioctl_pathset {
-        all_pathset.extend(pathset.iter().cloned());
-    }
-    if let Some(ref pathset) = policy.create_pathset {
-        all_pathset.extend(pathset.iter().cloned());
-    }
-    if let Some(ref pathset) = policy.delete_pathset {
-        all_pathset.extend(pathset.iter().cloned());
-    }
-    if let Some(ref pathset) = policy.rename_pathset {
-        all_pathset.extend(pathset.iter().cloned());
-    }
-    if let Some(ref pathset) = policy.symlink_pathset {
-        all_pathset.extend(pathset.iter().cloned());
-    }
-    if let Some(ref pathset) = policy.truncate_pathset {
-        all_pathset.extend(pathset.iter().cloned());
-    }
-    if let Some(ref pathset) = policy.readdir_pathset {
-        all_pathset.extend(pathset.iter().cloned());
-    }
-    if let Some(ref pathset) = policy.mkdir_pathset {
-        all_pathset.extend(pathset.iter().cloned());
-    }
-    if let Some(ref pathset) = policy.rmdir_pathset {
-        all_pathset.extend(pathset.iter().cloned());
-    }
-    if let Some(ref pathset) = policy.mkdev_pathset {
-        all_pathset.extend(pathset.iter().cloned());
-    }
-    if let Some(ref pathset) = policy.mkfifo_pathset {
-        all_pathset.extend(pathset.iter().cloned());
-    }
-    if let Some(ref pathset) = policy.bind_pathset {
-        all_pathset.extend(pathset.iter().cloned());
-    }
-
-    // Step 2: Accumulate access rights using the `all_pathset`.
-    let mut acl: HashMap<AccessFs, Vec<XPathBuf>, RandomState> = HashMap::default();
-    for path in all_pathset {
-        let mut access = AccessFs::EMPTY;
-
-        if policy
-            .read_pathset
-            .as_ref()
-            .map(|set| set.contains(&path))
-            .unwrap_or(false)
-        {
-            access |= AccessFs::ReadFile;
-        }
-        if policy
-            .write_pathset
-            .as_ref()
-            .map(|set| set.contains(&path))
-            .unwrap_or(false)
-        {
-            access |= AccessFs::WriteFile;
-        }
-        if policy
-            .exec_pathset
-            .as_ref()
-            .map(|set| set.contains(&path))
-            .unwrap_or(false)
-        {
-            access |= AccessFs::Execute;
-        }
-        if policy
-            .ioctl_pathset
-            .as_ref()
-            .map(|set| set.contains(&path))
-            .unwrap_or(false)
-        {
-            access |= AccessFs::IoctlDev;
-        }
-        if policy
-            .create_pathset
-            .as_ref()
-            .map(|set| set.contains(&path))
-            .unwrap_or(false)
-        {
-            access |= AccessFs::MakeReg;
-        }
-        if policy
-            .delete_pathset
-            .as_ref()
-            .map(|set| set.contains(&path))
-            .unwrap_or(false)
-        {
-            access |= AccessFs::RemoveFile;
-        }
-        if policy
-            .rename_pathset
-            .as_ref()
-            .map(|set| set.contains(&path))
-            .unwrap_or(false)
-        {
-            access |= AccessFs::Refer;
-        }
-        if policy
-            .symlink_pathset
-            .as_ref()
-            .map(|set| set.contains(&path))
-            .unwrap_or(false)
-        {
-            access |= AccessFs::MakeSym;
-        }
-        if policy
-            .truncate_pathset
-            .as_ref()
-            .map(|set| set.contains(&path))
-            .unwrap_or(false)
-        {
-            access |= AccessFs::Truncate;
-        }
-        if policy
-            .readdir_pathset
-            .as_ref()
-            .map(|set| set.contains(&path))
-            .unwrap_or(false)
-        {
-            access |= AccessFs::ReadDir;
-        }
-        if policy
-            .mkdir_pathset
-            .as_ref()
-            .map(|set| set.contains(&path))
-            .unwrap_or(false)
-        {
-            access |= AccessFs::MakeDir;
-        }
-        if policy
-            .rmdir_pathset
-            .as_ref()
-            .map(|set| set.contains(&path))
-            .unwrap_or(false)
-        {
-            access |= AccessFs::RemoveDir;
-        }
-        if policy
-            .mkdev_pathset
-            .as_ref()
-            .map(|set| set.contains(&path))
-            .unwrap_or(false)
-        {
-            access |= AccessFs::MakeChar;
-        }
-        if policy
-            .mkfifo_pathset
-            .as_ref()
-            .map(|set| set.contains(&path))
-            .unwrap_or(false)
-        {
-            access |= AccessFs::MakeFifo;
-        }
-        if policy
-            .bind_pathset
-            .as_ref()
-            .map(|set| set.contains(&path))
-            .unwrap_or(false)
-        {
-            access |= AccessFs::MakeSock;
-        }
-
-        if access.is_empty() {
-            continue;
-        }
-
-        acl.entry(access).or_insert_with(Vec::new).push(path);
-    }
-
-    // Step 3: Create ruleset and enter (access, path-set) pairs.
-    let mut ruleset = ruleset.create()?;
-    for (access, paths) in &acl {
-        ruleset = ruleset.add_rules(landlock_path_beneath_rules(paths, *access))?;
-    }
-
-    ruleset
-        .add_rules(
-            network_rules_bind
-                .into_iter()
-                .map(|port| Ok::<NetPort, RulesetError>(NetPort::new(port, AccessNet::BindTcp))),
-        )?
-        .add_rules(
-            network_rules_conn
-                .into_iter()
-                .map(|port| Ok::<NetPort, RulesetError>(NetPort::new(port, AccessNet::ConnectTcp))),
-        )?
-        .restrict_self()
-}
-
 /// Simple human size formatter.
 #[allow(clippy::arithmetic_side_effects)]
 #[allow(clippy::cast_precision_loss)]
@@ -2108,7 +1774,7 @@ pub(crate) fn validate_dev_null(fd: RawFd) -> Result<(), Errno> {
     // SAFETY: fd must be valid FD, let's trust the caller here.
     let fd = unsafe { BorrowedFd::borrow_raw(fd) };
 
-    let statx = fstatx(&fd, STATX_BASIC_STATS)?;
+    let statx = fstatx(fd, STATX_BASIC_STATS)?;
 
     // Basic checks: Is it a character device and does it match /dev/null?
     // Check file mode: stx_mode includes the file type bits.
@@ -2177,6 +1843,7 @@ pub fn check_fd_leaks(fd_max: Option<RawFd>) -> u32 {
         // Create a PathBuf from the string representation of the file descriptor
         let link_path = proc_fd_path.join(fd_str.into_owned()); // Convert Cow<str> into a String and then into a PathBuf
 
+        #[allow(clippy::disallowed_methods)]
         match std::fs::read_link(&link_path) {
             Ok(target_path) => {
                 eprintln!("!!! Leaked file descriptor {fd} -> {target_path:?} !!!");
@@ -2322,6 +1989,7 @@ impl FromStr for ExportMode {
 
 impl ExportMode {
     /// Return the export mode specified by the environment.
+    #[allow(clippy::disallowed_methods)]
     pub fn from_env() -> Option<ExportMode> {
         Self::from_str(&std::env::var(crate::config::ENV_DUMP_SCMP).ok()?).ok()
     }
