@@ -8,7 +8,7 @@
 
 use std::{
     io::Write,
-    os::fd::{AsRawFd, BorrowedFd, RawFd},
+    os::fd::{BorrowedFd, RawFd},
     time::{SystemTime, UNIX_EPOCH},
 };
 
@@ -208,7 +208,7 @@ pub(crate) struct LockedWriter<'a> {
 
 impl<'a> LockedWriter<'a> {
     pub(crate) fn new(fd: BorrowedFd<'a>) -> Result<Self, Errno> {
-        lock_fd(&fd, true, true)?;
+        lock_fd(fd, true, true)?;
         Ok(Self { fd })
     }
 }
@@ -216,7 +216,7 @@ impl<'a> LockedWriter<'a> {
 impl Drop for LockedWriter<'_> {
     fn drop(&mut self) {
         let _ = self.flush();
-        let _ = unlock_fd(&self.fd);
+        let _ = unlock_fd(self.fd);
     }
 }
 
@@ -229,7 +229,7 @@ impl Write for LockedWriter<'_> {
         // SAFETY:
         // 1. Retry on EINTR.
         // 2. Ignore EINVAL which indicates special file.
-        match retry_on_eintr(|| fdatasync(self.fd.as_raw_fd())) {
+        match retry_on_eintr(|| fdatasync(self.fd)) {
             Ok(_) | Err(Errno::EINVAL) => Ok(()),
             Err(e) => Err(std::io::Error::from_raw_os_error(e as i32)),
         }
@@ -270,8 +270,8 @@ pub fn log_init(default_level: LogLevel, default_log_fd: Option<RawFd>) -> Resul
     if !tty {
         if fd == Some(libc::STDERR_FILENO) && std::env::var_os(ENV_QUIET_TTY).is_none() {
             // check if we have a TTY
-            tty = nix::unistd::isatty(libc::STDIN_FILENO).unwrap_or(false)
-                && nix::unistd::isatty(libc::STDERR_FILENO).unwrap_or(false);
+            tty = nix::unistd::isatty(std::io::stdin()).unwrap_or(false)
+                && nix::unistd::isatty(std::io::stderr()).unwrap_or(false);
         } else {
             tty = false;
         }
@@ -336,8 +336,8 @@ pub fn log_init_simple(default_level: LogLevel) -> Result<(), Errno> {
     if !tty {
         if fd == libc::STDERR_FILENO && std::env::var_os(ENV_QUIET_TTY).is_none() {
             // check if we have a TTY
-            tty = nix::unistd::isatty(libc::STDIN_FILENO).unwrap_or(false)
-                && nix::unistd::isatty(libc::STDERR_FILENO).unwrap_or(false);
+            tty = nix::unistd::isatty(std::io::stdin()).unwrap_or(false)
+                && nix::unistd::isatty(std::io::stderr()).unwrap_or(false);
         } else {
             tty = false;
         }
@@ -350,8 +350,8 @@ pub fn log_init_simple(default_level: LogLevel) -> Result<(), Errno> {
     let mut tty = std::env::var_os(ENV_FORCE_TTY).is_some();
     if !tty {
         if std::env::var_os(ENV_QUIET_TTY).is_none() {
-            tty = nix::unistd::isatty(libc::STDIN_FILENO).unwrap_or(false)
-                && nix::unistd::isatty(libc::STDERR_FILENO).unwrap_or(false);
+            tty = nix::unistd::isatty(std::io::stdin()).unwrap_or(false)
+                && nix::unistd::isatty(std::io::stderr()).unwrap_or(false);
         } else {
             tty = false;
         }
