@@ -29,9 +29,9 @@ use memchr::{
 };
 use nix::{
     errno::Errno,
-    fcntl::{openat2, OFlag, OpenHow, ResolveFlag},
+    fcntl::{openat2, OFlag, OpenHow, ResolveFlag, AT_FDCWD},
     libc::pid_t,
-    unistd::{close, Pid},
+    unistd::Pid,
     NixPath,
 };
 use once_cell::sync::Lazy;
@@ -784,22 +784,12 @@ impl XPath {
 
     /// Check if path is the root path, ie `/`.
     pub fn is_rootfs(&self) -> bool {
-        self.as_bytes().iter().all(|b| *b == b'/')
-    }
-
-    /// Check if path points to devfs root dir, ie. `/dev`.
-    pub fn is_devfs(&self) -> bool {
-        const DEV_LEN: usize = b"/dev".len();
-        const DEV_DIR_LEN: usize = b"/dev/".len();
-
-        match self.len() {
-            DEV_LEN if self.is_equal(b"/dev") => true,
-            DEV_DIR_LEN if self.is_equal(b"/dev/") => true,
-            _ => false,
-        }
+        self.as_bytes().iter().all(|b| matches!(*b, b'/' | b'.'))
     }
 
     /// Check if path points to procfs root dir, ie. `/proc`.
+    ///
+    /// `self` must be canonicalized.
     pub fn is_procfs(&self) -> bool {
         const PROC_LEN: usize = b"/proc".len();
         const PROC_DIR_LEN: usize = b"/proc/".len();
@@ -811,44 +801,26 @@ impl XPath {
         }
     }
 
-    /// Check if path points to sysfs root dir, ie. `/sys`.
-    pub fn is_sysfs(&self) -> bool {
-        const SYS_LEN: usize = b"/sys".len();
-        const SYS_DIR_LEN: usize = b"/sys/".len();
-
-        match self.len() {
-            SYS_LEN if self.is_equal(b"/sys") => true,
-            SYS_DIR_LEN if self.is_equal(b"/sys/") => true,
-            _ => false,
-        }
-    }
-
     /// Check if path points to devfs, ie. starts with `/dev`.
     /// The literal path `/dev` returns false.
+    ///
+    /// `self` must be canonicalized.
     pub fn is_dev(&self) -> bool {
         self.starts_with(b"/dev/")
     }
 
     /// Check if path points to procfs, ie. starts with `/proc`.
     /// The literal path `/proc` returns false.
+    ///
+    /// `self` must be canonicalized.
     pub fn is_proc(&self) -> bool {
         self.starts_with(b"/proc/")
     }
 
-    /// Check if path points to sysfs, ie. starts with `/sys`.
-    /// The literal path `/sys` returns false.
-    pub fn is_sys(&self) -> bool {
-        self.starts_with(b"/sys/")
-    }
-
     /// Check if path points to a static path.
     /// See proc_init in config.rs
     pub fn is_static(&self) -> bool {
-        self.is_rootfs()
-            || self.is_devfs()
-            || self.is_procfs()
-            || self.is_sysfs()
-            || self.is_equal(b"/dev/null")
+        self.is_rootfs() || self.is_procfs() || self.is_equal(b"/dev/null")
     }
 
     /// Check if path points to per-process procfs directory, ie. starts with `/proc/$pid`.
@@ -890,8 +862,8 @@ impl XPath {
                 how.resolve(ResolveFlag::RESOLVE_NO_MAGICLINKS | ResolveFlag::RESOLVE_NO_SYMLINKS);
         }
 
-        retry_on_eintr(|| openat2(libc::AT_FDCWD, self, how))
-            .map(close)
+        retry_on_eintr(|| openat2(AT_FDCWD, self, how))
+            .map(drop)
             .is_ok()
     }
 
@@ -1012,7 +984,7 @@ impl XPath {
     }
 
     /// Create a new `XPath` from a byte slice.
-    pub fn from_bytes(slice: &[u8]) -> &XPath {
+    pub const fn from_bytes(slice: &[u8]) -> &XPath {
         // SAFETY: XPath has repr(transparent)
         unsafe { std::mem::transmute(slice) }
     }
