@@ -16,19 +16,19 @@
 //
 // SPDX-License-Identifier: GPL-3.0
 
+// SAFETY: This module has been liberated from unsafe code!
+// proc_set_at_secure_* tests call fork(2) therefore they need unsafe.
+#![cfg_attr(not(test), forbid(unsafe_code))]
+
 use std::{
     collections::{HashMap, HashSet},
     ffi::OsStr,
     fs::File,
     io::{BufReader, IoSlice, IoSliceMut, Read},
     ops::Range,
-    os::{
-        fd::{AsRawFd, RawFd},
-        unix::ffi::OsStrExt,
-    },
+    os::{fd::RawFd, unix::ffi::OsStrExt},
 };
 
-use ahash::RandomState;
 use btoi::{btoi, btoi_radix};
 use indexmap::IndexSet;
 use memchr::{
@@ -44,7 +44,7 @@ use nix::{
         sysinfo::sysinfo,
         uio::{process_vm_readv, process_vm_writev, RemoteIoVec},
     },
-    unistd::{lseek64, read, Gid, Pid, Uid, Whence},
+    unistd::{getpid, lseek64, read, Gid, Pid, Uid, Whence},
 };
 use procfs::{
     process::{MMPermissions, MMapPath, MemoryMaps},
@@ -56,8 +56,13 @@ use crate::{
     config::*,
     elf::ElfType,
     err::err2no,
-    fs::safe_open,
-    parsers::{map_result, parse_stat, parse_statm, parse_status, read_to_end},
+    fs::{parse_fd, safe_open},
+    hash::SydRandomState,
+    parsers::{
+        map_result,
+        proc::{parse_stat, parse_statm, parse_status},
+        read_to_end,
+    },
     path::XPathBuf,
     SydMemoryMap, SydSigSet, XPath,
 };
@@ -111,6 +116,10 @@ pub struct Status {
     pub sig_caught: SydSigSet,
 }
 
+// Safe `OFlags` to use when opening `/proc` files.
+const OFLAG_PROC: OFlag =
+    OFlag::from_bits_retain(libc::O_RDONLY | libc::O_CLOEXEC | libc::O_NOFOLLOW | libc::O_NOCTTY);
+
 // major 204 is a mess -- "Low-density serial ports"
 const LOW_DENSITY_NAMES: &[&str] = &[
     "LU0", "LU1", "LU2", "LU3", "FB0", "SA0", "SA1", "SA2", "SC0", "SC1", "SC2", "SC3", "FW0",
@@ -190,7 +199,7 @@ pub fn proc_stat(pid: Pid) -> Result<Stat, Errno> {
     let mut path = XPathBuf::from_pid(pid);
     path.push(b"stat");
 
-    let mut file = safe_open(Some(&PROC_FILE()), &path, OFlag::empty())
+    let mut file = safe_open(PROC_FILE(), &path, OFLAG_PROC)
         .map(File::from)
         .or(Err(Errno::ESRCH))?;
 
@@ -203,7 +212,7 @@ pub fn proc_statm(pid: Pid) -> Result<Statm, Errno> {
     let mut path = XPathBuf::from_pid(pid);
     path.push(b"statm");
 
-    let mut file = safe_open(Some(&PROC_FILE()), &path, OFlag::empty())
+    let mut file = safe_open(PROC_FILE(), &path, OFLAG_PROC)
         .map(File::from)
         .or(Err(Errno::ESRCH))?;
 
@@ -216,7 +225,7 @@ pub fn proc_status(pid: Pid) -> Result<Status, Errno> {
     let mut path = XPathBuf::from_pid(pid);
     path.push(b"status");
 
-    let mut file = safe_open(Some(&PROC_FILE()), &path, OFlag::empty())
+    let mut file = safe_open(PROC_FILE(), &path, OFLAG_PROC)
         .map(File::from)
         .or(Err(Errno::ESRCH))?;
 
@@ -262,7 +271,7 @@ pub fn proc_stack_pointer(pid: Pid) -> Result<u64, Errno> {
     let mut path = XPathBuf::from_pid(pid);
     path.push(b"syscall");
 
-    let mut file = safe_open(Some(&PROC_FILE()), &path, OFlag::empty())
+    let mut file = safe_open(PROC_FILE(), &path, OFLAG_PROC)
         .map(File::from)
         .map(BufReader::new)
         .or(Err(Errno::ESRCH))?;
@@ -290,11 +299,11 @@ pub fn proc_stack_pointer(pid: Pid) -> Result<u64, Errno> {
 }
 
 /// Get the auxiliary vector of the given `Pid`.
-pub fn proc_auxv(pid: Pid) -> Result<HashMap<u64, u64, RandomState>, Errno> {
+pub fn proc_auxv(pid: Pid) -> Result<HashMap<u64, u64, SydRandomState>, Errno> {
     let mut path = XPathBuf::from_pid(pid);
     path.push(b"auxv");
 
-    let mut file = safe_open(Some(&PROC_FILE()), &path, OFlag::empty())
+    let mut file = safe_open(PROC_FILE(), &path, OFLAG_PROC)
         .map(File::from)
         .or(Err(Errno::ESRCH))?;
 
@@ -330,7 +339,7 @@ pub fn proc_auxv(pid: Pid) -> Result<HashMap<u64, u64, RandomState>, Errno> {
 pub fn proc_cwd(pid: Pid) -> Result<XPathBuf, Errno> {
     let mut path = XPathBuf::from_pid(pid);
     path.push(b"cwd");
-    readlinkat(Some(PROC_FILE().as_raw_fd()), &path)
+    readlinkat(PROC_FILE(), &path)
         .map(XPathBuf::from)
         .or(Err(Errno::ESRCH))
 }
@@ -341,7 +350,7 @@ pub fn proc_comm(pid: Pid) -> Result<XPathBuf, Errno> {
     let mut path = XPathBuf::from_pid(pid);
     path.push(b"comm");
 
-    let mut file = safe_open(Some(&PROC_FILE()), &path, OFlag::empty())
+    let mut file = safe_open(PROC_FILE(), &path, OFLAG_PROC)
         .or(Err(Errno::ESRCH))
         .map(File::from)?;
 
@@ -379,7 +388,7 @@ pub fn proc_cmdline(pid: Pid) -> Result<XPathBuf, Errno> {
     let mut path = XPathBuf::from_pid(pid);
     path.push(b"cmdline");
 
-    let mut file = safe_open(Some(&PROC_FILE()), &path, OFlag::O_RDONLY)
+    let mut file = safe_open(PROC_FILE(), &path, OFLAG_PROC)
         .or(Err(Errno::ESRCH))
         .map(File::from)?;
 
@@ -434,7 +443,7 @@ pub fn proc_maps(pid: Pid) -> Result<Vec<SydMemoryMap>, Errno> {
     let mut pfd = XPathBuf::from_pid(pid);
     pfd.push(b"maps");
 
-    let reader = safe_open(Some(&PROC_FILE()), &pfd, OFlag::empty())
+    let reader = safe_open(PROC_FILE(), &pfd, OFLAG_PROC)
         .map(File::from)
         .map(BufReader::new)?;
 
@@ -448,7 +457,7 @@ pub fn proc_smaps(pid: Pid) -> Result<Vec<SydMemoryMap>, Errno> {
     let mut path = XPathBuf::from_pid(pid);
     path.push(b"smaps");
 
-    let reader = safe_open(Some(&PROC_FILE()), &path, OFlag::empty())
+    let reader = safe_open(PROC_FILE(), &path, OFLAG_PROC)
         .map(File::from)
         .map(BufReader::new)
         .or(Err(Errno::ESRCH))?;
@@ -475,20 +484,19 @@ pub fn proc_mmap(pid: Pid, addr: u64) -> Result<SydMemoryMap, Errno> {
 }
 
 /// Retrieve the system pipe max limit.
-pub fn proc_pipemax() -> Result<nix::libc::c_int, Errno> {
+pub fn proc_pipemax() -> Result<libc::c_int, Errno> {
     let fd = safe_open(
-        Some(&PROC_FILE()),
+        PROC_FILE(),
         XPath::from_bytes(b"sys/fs/pipe-max-size"),
-        OFlag::empty(),
-    )
-    .or(Err(Errno::ESRCH))?;
+        OFLAG_PROC,
+    )?;
 
     // Read up to 24 bytes.
     let mut data = [0u8; 24];
     let mut nread = 0;
     while nread < data.len() {
         #[allow(clippy::arithmetic_side_effects)]
-        match read(fd.as_raw_fd(), &mut data[nread..]) {
+        match read(&fd, &mut data[nread..]) {
             Ok(0) => break,
             Ok(n) => nread += n,
             Err(Errno::EINTR) => continue,
@@ -496,24 +504,23 @@ pub fn proc_pipemax() -> Result<nix::libc::c_int, Errno> {
         }
     }
 
-    btoi::<nix::libc::c_int>(&data[..nread]).or(Err(Errno::EINVAL))
+    btoi::<libc::c_int>(&data[..nread]).or(Err(Errno::EINVAL))
 }
 
 /// Retrieves the value of /proc/sys/fs/file-max.
 pub fn proc_fs_file_max() -> Result<u64, Errno> {
     let fd = safe_open(
-        Some(&PROC_FILE()),
+        PROC_FILE(),
         XPath::from_bytes(b"sys/fs/file-max"),
-        OFlag::empty(),
-    )
-    .or(Err(Errno::ESRCH))?;
+        OFLAG_PROC,
+    )?;
 
     // Read up to 24 bytes + 1 bytes for \n.
     let mut data = [0u8; 25];
     let mut nread = 0;
     while nread < data.len() {
         #[allow(clippy::arithmetic_side_effects)]
-        match read(fd.as_raw_fd(), &mut data[nread..]) {
+        match read(&fd, &mut data[nread..]) {
             Ok(0) => break,
             Ok(n) => nread += n,
             Err(Errno::EINTR) => continue,
@@ -528,18 +535,17 @@ pub fn proc_fs_file_max() -> Result<u64, Errno> {
 #[allow(clippy::disallowed_methods)]
 pub fn proc_mmap_min_addr() -> Result<u64, Errno> {
     let fd = safe_open(
-        Some(&PROC_FILE()),
+        PROC_FILE(),
         XPath::from_bytes(b"sys/vm/mmap_min_addr"),
-        OFlag::empty(),
-    )
-    .or(Err(Errno::ESRCH))?;
+        OFLAG_PROC,
+    )?;
 
     // Read up to 24 bytes + 1 bytes for \n.
     let mut data = [0u8; 25];
     let mut nread = 0;
     while nread < data.len() {
         #[allow(clippy::arithmetic_side_effects)]
-        match read(fd.as_raw_fd(), &mut data[nread..]) {
+        match read(&fd, &mut data[nread..]) {
             Ok(0) => break,
             Ok(n) => nread += n,
             Err(Errno::EINTR) => continue,
@@ -584,39 +590,39 @@ pub fn proc_tty(pid: Pid) -> Result<XPathBuf, Errno> {
         ))),
         4 => {
             if minor < 64 {
-                Ok(XPathBuf::from(format!("/dev/tty{}", minor)))
+                Ok(XPathBuf::from(format!("/dev/tty{minor}")))
             } else {
                 Ok(XPathBuf::from(format!("/dev/ttyS{}", minor - 64)))
             }
         }
-        11 => Ok(XPathBuf::from(format!("/dev/ttyB{}", minor))),
-        14 => Ok(XPathBuf::from(format!("/dev/tty{}", minor))), // Standard TTYs
-        17 => Ok(XPathBuf::from(format!("/dev/ttyH{}", minor))),
-        19 | 22 | 23 => Ok(XPathBuf::from(format!("/dev/ttyD{}", minor))),
-        24 => Ok(XPathBuf::from(format!("/dev/ttyE{}", minor))),
-        32 => Ok(XPathBuf::from(format!("/dev/ttyX{}", minor))),
-        43 => Ok(XPathBuf::from(format!("/dev/ttyI{}", minor))),
-        46 => Ok(XPathBuf::from(format!("/dev/ttyR{}", minor))),
-        48 => Ok(XPathBuf::from(format!("/dev/ttyL{}", minor))),
-        57 => Ok(XPathBuf::from(format!("/dev/ttyP{}", minor))),
-        71 => Ok(XPathBuf::from(format!("/dev/ttyF{}", minor))),
-        75 => Ok(XPathBuf::from(format!("/dev/ttyW{}", minor))),
-        78 | 112 => Ok(XPathBuf::from(format!("/dev/ttyM{}", minor))),
-        105 => Ok(XPathBuf::from(format!("/dev/ttyV{}", minor))),
+        11 => Ok(XPathBuf::from(format!("/dev/ttyB{minor}"))),
+        14 => Ok(XPathBuf::from(format!("/dev/tty{minor}"))), // Standard TTYs
+        17 => Ok(XPathBuf::from(format!("/dev/ttyH{minor}"))),
+        19 | 22 | 23 => Ok(XPathBuf::from(format!("/dev/ttyD{minor}"))),
+        24 => Ok(XPathBuf::from(format!("/dev/ttyE{minor}"))),
+        32 => Ok(XPathBuf::from(format!("/dev/ttyX{minor}"))),
+        43 => Ok(XPathBuf::from(format!("/dev/ttyI{minor}"))),
+        46 => Ok(XPathBuf::from(format!("/dev/ttyR{minor}"))),
+        48 => Ok(XPathBuf::from(format!("/dev/ttyL{minor}"))),
+        57 => Ok(XPathBuf::from(format!("/dev/ttyP{minor}"))),
+        71 => Ok(XPathBuf::from(format!("/dev/ttyF{minor}"))),
+        75 => Ok(XPathBuf::from(format!("/dev/ttyW{minor}"))),
+        78 | 112 => Ok(XPathBuf::from(format!("/dev/ttyM{minor}"))),
+        105 => Ok(XPathBuf::from(format!("/dev/ttyV{minor}"))),
         136..=143 => Ok(XPathBuf::from(format!(
             "/dev/pts/{}",
             minor + (major - 136) * 256
         ))),
-        148 => Ok(XPathBuf::from(format!("/dev/ttyT{}", minor))),
+        148 => Ok(XPathBuf::from(format!("/dev/ttyT{minor}"))),
         154 | 156 => Ok(XPathBuf::from(format!(
             "/dev/ttySR{}",
             minor + if major == 156 { 256 } else { 0 }
         ))),
-        164 => Ok(XPathBuf::from(format!("/dev/ttyCH{}", minor))),
-        166 => Ok(XPathBuf::from(format!("/dev/ttyACM{}", minor))),
-        172 => Ok(XPathBuf::from(format!("/dev/ttyMX{}", minor))),
-        174 => Ok(XPathBuf::from(format!("/dev/ttySI{}", minor))),
-        188 => Ok(XPathBuf::from(format!("/dev/ttyUSB{}", minor))),
+        164 => Ok(XPathBuf::from(format!("/dev/ttyCH{minor}"))),
+        166 => Ok(XPathBuf::from(format!("/dev/ttyACM{minor}"))),
+        172 => Ok(XPathBuf::from(format!("/dev/ttyMX{minor}"))),
+        174 => Ok(XPathBuf::from(format!("/dev/ttySI{minor}"))),
+        188 => Ok(XPathBuf::from(format!("/dev/ttyUSB{minor}"))),
         204 => {
             if minor as usize >= LOW_DENSITY_NAMES.len() {
                 Err(Errno::ENXIO)
@@ -627,12 +633,12 @@ pub fn proc_tty(pid: Pid) -> Result<XPathBuf, Errno> {
                 )))
             }
         }
-        208 => Ok(XPathBuf::from(format!("/dev/ttyU{}", minor))),
-        216 => Ok(XPathBuf::from(format!("/dev/ttyUB{}", minor))),
-        224 => Ok(XPathBuf::from(format!("/dev/ttyY{}", minor))),
-        227 => Ok(XPathBuf::from(format!("/dev/3270/tty{}", minor))),
-        229 => Ok(XPathBuf::from(format!("/dev/iseries/vtty{}", minor))),
-        256 => Ok(XPathBuf::from(format!("/dev/ttyEQ{}", minor))),
+        208 => Ok(XPathBuf::from(format!("/dev/ttyU{minor}"))),
+        216 => Ok(XPathBuf::from(format!("/dev/ttyUB{minor}"))),
+        224 => Ok(XPathBuf::from(format!("/dev/ttyY{minor}"))),
+        227 => Ok(XPathBuf::from(format!("/dev/3270/tty{minor}"))),
+        229 => Ok(XPathBuf::from(format!("/dev/iseries/vtty{minor}"))),
+        256 => Ok(XPathBuf::from(format!("/dev/ttyEQ{minor}"))),
         _ => Err(Errno::ENXIO),
     }
 }
@@ -666,7 +672,7 @@ pub fn proc_task_limit(pid: Pid, max: u64) -> Result<bool, Errno> {
     // Hence no concurrent readdir() is possible here.
     // Rewind the directory and get to work!
     let fd = PROC_FILE();
-    lseek64(fd.as_raw_fd(), 0, Whence::SeekSet)?;
+    lseek64(fd, 0, Whence::SeekSet)?;
 
     let this = Pid::this().as_raw();
 
@@ -760,7 +766,11 @@ pub fn proc_fd(
     let path_without_proc = &path[PROC.len()..];
     let next_slash_index = memchr(b'/', path_without_proc).unwrap_or(path_without_proc.len());
     let pid_section = &path_without_proc[..next_slash_index];
-    if pid_section.is_empty() || !pid_section.iter().all(|c| c.is_ascii_digit()) {
+    if pid_section
+        .first()
+        .map(|c| !c.is_ascii_digit())
+        .unwrap_or(true)
+    {
         // Skip /proc/not-a-pid paths.
         return Ok(None);
     }
@@ -809,7 +819,7 @@ pub fn proc_fd(
         })
     } else if is_equal(remaining_path, b"/cwd") {
         // Magic CWD.
-        Ok(Some((nix::libc::AT_FDCWD, pid)))
+        Ok(Some((libc::AT_FDCWD, pid)))
     } else if is_equal(remaining_path, b"/root") {
         // Magic ROOT.
         Ok(Some((-1, pid)))
@@ -892,7 +902,8 @@ pub fn proc_mem_limit(pid: Pid, max: u64) -> Result<bool, Errno> {
 ///
 /// A `Result` containing a vector of `(XPathBuf,u64,i32,i32)` objects
 /// for all unique paths with executable permissions along with their
-/// device ID and inode. On error, it returns an `Errno`.
+/// device ID and inode. On error, it returns an `Errno`. If no executable
+/// is found in maps, returns `Errno::ENOEXEC`.
 #[allow(clippy::type_complexity)]
 pub fn proc_executables(pid: Pid) -> Result<Vec<(XPathBuf, u64, i32, i32)>, Errno> {
     let maps = proc_maps(pid)?;
@@ -913,6 +924,11 @@ pub fn proc_executables(pid: Pid) -> Result<Vec<(XPathBuf, u64, i32, i32)>, Errn
         }
     }
 
+    // Check for no match.
+    if paths.is_empty() {
+        return Err(Errno::ENOEXEC);
+    }
+
     // Convert the IndexSet to a vector and return.
     Ok(paths.into_iter().collect())
 }
@@ -1043,17 +1059,17 @@ pub fn proc_set_at_secure(pid: Pid, elf_type: ElfType) -> Result<(), Errno> {
     // We also ensure no other key is ever present in this sequence.
     // The AT_NULL key indicates the end of the vector.
     #[allow(clippy::cast_possible_truncation)]
-    const AT_NULL: usize = nix::libc::AT_NULL as usize;
+    const AT_NULL: usize = libc::AT_NULL as usize;
     #[allow(clippy::cast_possible_truncation)]
-    const AT_UID: usize = nix::libc::AT_UID as usize;
+    const AT_UID: usize = libc::AT_UID as usize;
     #[allow(clippy::cast_possible_truncation)]
-    const AT_EUID: usize = nix::libc::AT_EUID as usize;
+    const AT_EUID: usize = libc::AT_EUID as usize;
     #[allow(clippy::cast_possible_truncation)]
-    const AT_GID: usize = nix::libc::AT_GID as usize;
+    const AT_GID: usize = libc::AT_GID as usize;
     #[allow(clippy::cast_possible_truncation)]
-    const AT_EGID: usize = nix::libc::AT_EGID as usize;
+    const AT_EGID: usize = libc::AT_EGID as usize;
     #[allow(clippy::cast_possible_truncation)]
-    const AT_SECURE: usize = nix::libc::AT_SECURE as usize;
+    const AT_SECURE: usize = libc::AT_SECURE as usize;
     #[allow(clippy::cast_possible_truncation)]
     const AT_REQKEY: &[usize] = &[AT_UID, AT_EUID, AT_GID, AT_EGID, AT_SECURE];
     let mut required_index = 0;
@@ -1158,6 +1174,60 @@ pub fn proc_set_at_secure(pid: Pid, elf_type: ElfType) -> Result<(), Errno> {
     Err(Errno::ENOENT)
 }
 
+/// List `/proc/pid/fd` contents.
+///
+/// Return a vector of `(RawFd, XPathBuf)` tuples, where each `RawFd`
+/// is the file descriptor number and the `XPathBuf` is the path it points to.
+///
+/// Useful for debugging file descriptor leaks.
+#[allow(clippy::type_complexity)]
+pub fn proc_pid_fd(pid: Option<Pid>) -> Result<Vec<(RawFd, XPathBuf)>, Errno> {
+    let pid = pid.unwrap_or_else(getpid);
+
+    let mut dir = XPathBuf::from_pid(pid);
+    dir.push(b"fd");
+
+    let dir = safe_open(PROC_FILE(), &dir, OFLAG_PROC | OFlag::O_DIRECTORY)?;
+
+    let mut dot = 0u8;
+    let mut res = vec![];
+    loop {
+        let mut entries = match getdents64(&dir, DIRENT_BUF_SIZE) {
+            Ok(entries) => entries,
+            Err(Errno::ECANCELED) => break, // EOF or empty directory
+            Err(errno) => return Err(errno),
+        };
+
+        for entry in &mut entries {
+            #[allow(clippy::arithmetic_side_effects)]
+            if dot < 2 && entry.is_dot() {
+                dot += 1;
+                continue;
+            }
+            let fd = parse_fd(XPath::from_bytes(entry.name_bytes()))?;
+            let target = readlinkat(&dir, entry.name_bytes()).map(XPathBuf::from)?;
+
+            res.push((fd, target));
+        }
+    }
+
+    Ok(res)
+}
+
+/// Log `/proc/self/fd` contents.
+///
+/// Useful for debugging file descriptor leaks.
+pub fn log_proc_pid_fd(pid: Option<Pid>) -> Result<(), Errno> {
+    let fds = proc_pid_fd(pid)?;
+    let pid = pid.unwrap_or_else(getpid);
+
+    crate::error!("ctx": "log_proc_self_fd",
+        "msg": format!("List of /proc/{pid}/fd"),
+        "fds": fds, "pid": pid.as_raw());
+
+    Ok(())
+}
+
 #[cfg(test)]
 mod tests {
     use std::{io::Write, os::unix::process::CommandExt, process::Command};
@@ -1230,7 +1300,7 @@ mod tests {
 
         assert_eq!(
             proc_fd(this, &xpath!("/proc/{this}/cwd"), true),
-            Ok(Some((nix::libc::AT_FDCWD, this)))
+            Ok(Some((libc::AT_FDCWD, this)))
         );
         assert_eq!(
             proc_fd(this, &xpath!("/proc/{this}/exe"), true),
