@@ -58,7 +58,7 @@ use std::{
     io::{BufWriter, Write},
     net::{IpAddr, SocketAddrV4, SocketAddrV6},
     os::{
-        fd::{AsRawFd, RawFd},
+        fd::AsRawFd,
         unix::{ffi::OsStrExt, fs::OpenOptionsExt, net::UnixStream, process::CommandExt},
     },
     path::Path,
@@ -66,7 +66,6 @@ use std::{
     str::FromStr,
 };
 
-use ahash::RandomState;
 use libseccomp::{scmp_cmp, ScmpAction, ScmpFilterContext, ScmpSyscall};
 use nix::{
     errno::Errno,
@@ -77,7 +76,7 @@ use nix::{
         wait::{Id, WaitPidFlag},
     },
     time::{clock_gettime, ClockId},
-    unistd::{close, fork, getgid, getpid, getuid, sethostname, ForkResult},
+    unistd::{fork, getgid, getpid, getuid, sethostname, ForkResult},
     NixPath,
 };
 use sendfd::SendWithFd;
@@ -87,16 +86,18 @@ use syd::{
     config::*,
     err::{err2no, SydResult},
     error,
-    fs::set_cloexec,
+    fs::{pidfd_open, set_cloexec},
+    hash::SydRandomState,
     hook::Supervisor,
     info,
+    landlock_policy::LandlockPolicy,
     log::log_init,
     nsflags_name,
     path::XPathBuf,
     sandbox::Sandbox,
     syslog::LogLevel,
     unshare::{GidMap, UidMap},
-    warn, ExportMode, LandlockPolicy,
+    warn, ExportMode,
 };
 
 // Set global allocator to tcmalloc if profiling is enabled.
@@ -197,9 +198,9 @@ fn main() -> SydResult<ExitCode> {
     let mut sandbox: Sandbox = Sandbox::default();
     let mut cmd_arg0: Option<OsString> = None;
     let mut cmd_argv: Vec<OsString> = vec![];
-    let mut cmd_envp: HashSet<OsString, RandomState> = HashSet::default();
+    let mut cmd_envp: HashSet<OsString, SydRandomState> = HashSet::default();
 
-    // SYD_PROXY_{HOST,PORT} -> proxy/ext/{host,port}
+    // SYD_PROXY_{HOST,PORT,UNIX} -> proxy/ext/{host,port,unix}
     #[allow(clippy::disallowed_methods)]
     match env::var(ENV_PROXY_HOST) {
         Ok(host) => sandbox
@@ -216,6 +217,14 @@ fn main() -> SydResult<ExitCode> {
         Err(env::VarError::NotPresent) => {}
         Err(error) => panic!("Invalid UTF-8 in {ENV_PROXY_PORT}: {error}"),
     };
+    #[allow(clippy::disallowed_methods)]
+    match env::var(ENV_PROXY_UNIX) {
+        Ok(unix) => sandbox
+            .config(&format!("proxy/ext/unix:{unix}"))
+            .expect(ENV_PROXY_UNIX),
+        Err(env::VarError::NotPresent) => {}
+        Err(error) => panic!("Invalid UTF-8 in {ENV_PROXY_UNIX}: {error}"),
+    };
 
     // Initialize Options.
     let mut user_parse = false;
@@ -414,6 +423,9 @@ fn main() -> SydResult<ExitCode> {
     if sandbox.unshare_pid() {
         sandbox.set_unshare_mount(true);
     }
+    if sandbox.unshare_mount() {
+        sandbox.set_unshare_pid(true);
+    }
 
     let proxy_debug = env::var_os("SYD_TOR_DEBUG").is_some();
     let proxy = if sandbox.has_proxy() {
@@ -421,17 +433,7 @@ fn main() -> SydResult<ExitCode> {
         sandbox.set_unshare_net(true);
 
         // Step 1: Create a PIDFd of this process.
-        // SAFETY: nix has no wrapper for pidfd_open.
-        let pidfd = match unsafe {
-            nix::libc::syscall(
-                nix::libc::SYS_pidfd_open,
-                getpid().as_raw(),
-                nix::libc::PIDFD_NONBLOCK,
-            )
-        } {
-            e if e < 0 => return Err(Errno::last().into()),
-            fd => fd as RawFd,
-        };
+        let pidfd = pidfd_open(getpid(), libc::PIDFD_NONBLOCK)?;
 
         // Step 2: Create a UNIX socket pair.
         let (stream_parent, stream_child) = UnixStream::pair()?;
@@ -439,8 +441,7 @@ fn main() -> SydResult<ExitCode> {
         // Step 3: Unset the CLOEXEC flags on the file descriptors.
         // PIDFds and Rust sockets are automatically CLOEXEC.
         set_cloexec(&pidfd, false)?;
-        let fd = stream_parent.as_raw_fd(); // Borrows FD.
-        set_cloexec(&fd, false)?;
+        set_cloexec(&stream_parent, false)?;
 
         // Step 4: Prepare environment of the syd-tor process.
         // Filter the environment variables to only include the list below:
@@ -448,7 +449,7 @@ fn main() -> SydResult<ExitCode> {
         // 2. LD_LIBRARY_PATH
         // 3. SYD_TOR_DEBUG
         // 4. SYD_TOR_RULES
-        let safe_env: HashMap<_, _, RandomState> = env::vars_os()
+        let safe_env: HashMap<_, _, SydRandomState> = env::vars_os()
             .filter(|(key, _)| {
                 matches!(
                     key.as_bytes(),
@@ -465,12 +466,17 @@ fn main() -> SydResult<ExitCode> {
         if proxy_debug {
             cmd.arg("-d");
         }
-        cmd.arg(format!("-p{pidfd}"));
-        cmd.arg(format!("-i{fd}"));
-        cmd.arg(format!(
-            "-o{}:{}",
-            sandbox.proxy_ext_addr, sandbox.proxy_ext_port
-        ));
+        cmd.arg(format!("-p{}", pidfd.as_raw_fd()));
+        cmd.arg(format!("-i{}", stream_parent.as_raw_fd()));
+        // proxy/ext/unix has precedence over proxy/ext/host.
+        if let Some(ref proxy_ext_unix) = sandbox.proxy_ext_unix {
+            cmd.arg(format!("-o{proxy_ext_unix}"));
+        } else {
+            cmd.arg(format!(
+                "-o{}:{}",
+                sandbox.proxy_ext_addr, sandbox.proxy_ext_port
+            ));
+        }
         //cmd.arg(format!("-b{}", PIPE_BUF));
         cmd.env_clear();
         cmd.envs(safe_env);
@@ -482,7 +488,7 @@ fn main() -> SydResult<ExitCode> {
             cmd.stderr(Stdio::null());
         }
         cmd.spawn()?;
-        close(pidfd)?;
+        drop(pidfd);
         if proxy_debug {
             warn!("ctx": "spawn_proxy",
                 "msg": format!("proxy is now forwarding external traffic to {}!{}",
@@ -611,7 +617,8 @@ fn main() -> SydResult<ExitCode> {
             }
             Err(errno) => {
                 error!("ctx": "bring_up_loopback_device",
-                    "error": format!("bring up loopback network device error: {errno}"));
+                    "msg": format!("bring up loopback network device error: {errno}"),
+                    "err": errno as i32);
             }
         }
 
@@ -658,7 +665,9 @@ fn main() -> SydResult<ExitCode> {
                                 warn!("ctx": "send_proxy_fd",
                                     "msg": format!("proxy fd {} sent to syd-tor, IPv{ipv} traffic forwarding is now started \\o/",
                                         lfd.as_raw_fd()));
-                                warn!("ctx": "send_proxy_fd", "syd": "ping", "msg": "Change return success. Going and coming without error.");
+                                warn!("ctx": "send_proxy_fd",
+                                    "syd": "ping",
+                                    "msg": "Change return success. Going and coming without error.");
                             } else {
                                 info!("ctx": "send_proxy_fd",
                                     "msg": format!("proxy fd {} sent to syd-tor, IPv{ipv} traffic forwarding is now started \\o/",
@@ -668,14 +677,16 @@ fn main() -> SydResult<ExitCode> {
                         }
                         Err(error) => {
                             error!("ctx": "send_proxy_fd",
-                                "error": format!("proxy fd {} send to syd-tor error: {}, traffic will not be forwarded",
-                                    lfd.as_raw_fd(), err2no(&error)));
+                                "msg": format!("proxy fd {} send to syd-tor error: {}, traffic will not be forwarded",
+                                    lfd.as_raw_fd(), err2no(&error)),
+                                "err": err2no(&error) as i32);
                         }
                     }
                 }
                 Err(errno) => {
                     error!("ctx": "bind_proxy",
-                        "error": format!("proxy bind to IPv{ipv} {proxy_addr}!{proxy_port} error: {errno}"));
+                        "msg": format!("proxy bind to IPv{ipv} {proxy_addr}!{proxy_port} error: {errno}"),
+                        "err": errno as i32);
                 }
             }
             drop(stream_child);
@@ -717,7 +728,7 @@ fn main() -> SydResult<ExitCode> {
         ForkResult::Child => {
             let retval = match Supervisor::run(sandbox, &argv0, cmd_argv, Some(&cmd_envp), cmd_arg0)
             {
-                Ok(retval) => nix::libc::c_int::from(retval),
+                Ok(retval) => libc::c_int::from(retval),
                 Err(error) => {
                     let errno = Errno::last();
                     eprintln!("{error:?}");
@@ -725,7 +736,7 @@ fn main() -> SydResult<ExitCode> {
                 }
             };
             // SAFETY: _exit is safe to call in child.
-            unsafe { nix::libc::_exit(retval) };
+            unsafe { libc::_exit(retval) };
         }
         ForkResult::Parent { child, .. } => {
             // Ensure we release all unneeded resources.
@@ -735,10 +746,10 @@ fn main() -> SydResult<ExitCode> {
             drop(sandbox);
             // SAFETY: nix does not have a close_range wrapper yet.
             let _ = unsafe {
-                nix::libc::syscall(
-                    nix::libc::SYS_close_range,
-                    nix::libc::STDERR_FILENO + 1,
-                    nix::libc::c_int::MAX,
+                libc::syscall(
+                    libc::SYS_close_range,
+                    libc::STDERR_FILENO + 1,
+                    libc::c_int::MAX,
                     0,
                 )
             };
@@ -755,7 +766,7 @@ fn main() -> SydResult<ExitCode> {
 
         ..Default::default()
     };
-    let _ = syd::landlock_operation(abi, &policy);
+    let _ = policy.restrict_self(abi);
 
     // SAFETY: Set up a seccomp filter which only allows
     // 1. write to standard error.
@@ -801,13 +812,13 @@ fn main() -> SydResult<ExitCode> {
             filter.add_rule_conditional(
                 ScmpAction::Allow,
                 syscall,
-                &[scmp_cmp!($arg0 == nix::libc::STDERR_FILENO as u64)],
+                &[scmp_cmp!($arg0 == libc::STDERR_FILENO as u64)],
             )?;
         }
     }
 
     // Prevent executable memory.
-    const PROT_EXEC: u64 = nix::libc::PROT_EXEC as u64;
+    const PROT_EXEC: u64 = libc::PROT_EXEC as u64;
     for name in ["mmap", "mmap2", "mprotect"] {
         #[allow(clippy::disallowed_methods)]
         filter.add_rule_conditional(
