@@ -507,6 +507,17 @@ macro_rules! skip_unless_pty {
     }};
 }
 
+#[macro_export]
+macro_rules! skip_unless_at_execve_check_is_supported {
+    () => {{
+        if !*HAVE_AT_EXECVE_CHECK {
+            eprintln!("execveat(2) flag AT_EXECVE_CHECK is not supported, skipping!");
+            std::env::set_var("SYD_TEST_SOFT_FAIL", "1");
+            return Ok(());
+        }
+    }};
+}
+
 /// Holds a `String` to run `syd`.
 /// Honours CARGO_BIN_EXE_syd environment variable.
 pub static SYD: Lazy<String> =
@@ -604,6 +615,11 @@ pub static SYD_DO: Lazy<String> = Lazy::new(|| match env::var("CARGO_BIN_EXE_syd
     .expect("syd-test-do"),
 });
 
+/// Holds a `String` to run `syd-x`.
+/// Honours CARGO_BIN_EXE_syd-x environment variable.
+pub static SYD_X: Lazy<String> =
+    Lazy::new(|| env::var("CARGO_BIN_EXE_syd-x").unwrap_or("syd-x".to_string()));
+
 /// A boolean which specifies if we're running under SourceHut CI.
 pub static CI_BUILD: Lazy<bool> = Lazy::new(|| env::var("JOB_ID").ok().is_some());
 
@@ -1119,7 +1135,7 @@ pub fn format_duration(d: Duration) -> String {
     let minutes = (total_seconds % 3600) / 60;
     let seconds = total_seconds % 60;
 
-    format!("{}h {}m {}s", hours, minutes, seconds)
+    format!("{hours}h {minutes}m {seconds}s")
 }
 
 /// Checks if the current running binary is 32-bit and the host system is 64-bit.
