@@ -3,7 +3,7 @@
 use std::{
     env, mem,
     os::{
-        fd::{AsRawFd, RawFd},
+        fd::{AsRawFd, FromRawFd, OwnedFd, RawFd},
         unix::ffi::OsStrExt,
     },
     ptr,
@@ -14,7 +14,6 @@ use libloading::os::unix::Symbol;
 use nix::{
     errno::Errno,
     libc,
-    sched::{unshare, CloneFlags},
     sys::signal::{kill, sigprocmask, SigSet, SigmaskHow, Signal},
     unistd::{close, getpid, read, write},
 };
@@ -22,7 +21,7 @@ use nix::{
 use crate::{
     caps,
     config::*,
-    fs::{duprand, set_cloexec},
+    fs::retry_on_eintr,
     safe_drop_cap,
     unshare::{
         error::ErrorCode as Err,
@@ -41,7 +40,6 @@ unsafe fn fail_errno(code: Err, errno: i32) -> ! {
         Err::Seccomp => c"syd: seccomp error".as_ptr(),
         Err::SeccompSendFd => c"syd: seccomp send notify-fd error".as_ptr(),
         Err::SeccompWaitFd => c"syd: seccomp wait for notify-fd error".as_ptr(),
-        Err::UnshareFiles => c"syd: error unsharing files".as_ptr(),
         #[cfg(any(target_arch = "x86", target_arch = "x86_64"))]
         Err::SetTSC => c"syd: set-tsc error".as_ptr(),
     };
@@ -65,9 +63,39 @@ macro_rules! fail_errno_safe {
 
 #[allow(clippy::cognitive_complexity)]
 pub extern "C" fn child_after_clone(arg: *mut libc::c_void) -> libc::c_int {
+    // SAFETY: Do not leak the static file descriptors to the sandbox process.
+    proc_close();
+
+    // SAFETY: Do not leak the log file descriptor to the sandbox process.
+    let log_fd = if let Some(fd) = env::var_os(ENV_LOG_FD) {
+        btoi::<RawFd>(fd.as_bytes()).ok()
+    } else {
+        None
+    };
+    if let Some(log_fd) = log_fd {
+        let _ = close(log_fd);
+    }
+
+    // SAFETY: Clean Syd environment variables from process environment.
+    // Note, we have just used ENV_LOG_FD above and do not need it anymore.
+    for (key, _) in env::vars_os() {
+        if key.as_bytes().starts_with(b"SYD_") && !key.as_bytes().starts_with(b"SYD_TEST_") {
+            env::remove_var(key);
+        }
+    }
+
     // SAFETY: arg is a valid ChildInfo structure.
     let child: Box<ChildInfo> = unsafe { Box::from_raw(arg as *mut ChildInfo) };
 
+    // We'll write seccomp notify fd to the second pipe,
+    // and read the acknowledgement notification from
+    // the first pipe.
+    let (pipe_ro, pipe_rw) = (child.seccomp_pipefd.0 .0, child.seccomp_pipefd.1 .1);
+
+    // Close the unused ends of the pipes.
+    drop(child.seccomp_pipefd.0 .1);
+    drop(child.seccomp_pipefd.1 .0);
+
     if let Some(&sig) = child.cfg.death_sig.as_ref() {
         if let Err(errno) = Errno::result(unsafe {
             libc::prctl(libc::PR_SET_PDEATHSIG, sig as libc::c_ulong, 0, 0, 0)
@@ -119,13 +147,7 @@ pub extern "C" fn child_after_clone(arg: *mut libc::c_void) -> libc::c_int {
         }
     }
 
-    let fdpair = if let Some(seccomp_filter) = &child.seccomp_filter {
-        // We'll write seccomp notify fd to the second pipe,
-        // and read the acknowledgement notification from
-        // the first pipe.
-        let pipe_ro = &child.seccomp_pipefd.0 .0;
-        let pipe_rw = &child.seccomp_pipefd.1 .1;
-
+    if let Some(seccomp_filter) = child.seccomp_filter {
         // Load the seccomp filter.
         if let Err(scmp_err) = seccomp_filter.load() {
             fail_errno_safe!(
@@ -139,19 +161,10 @@ pub extern "C" fn child_after_clone(arg: *mut libc::c_void) -> libc::c_int {
         }
 
         // Get seccomp notification fd.
-        // SAFETY: ScmpFilterContext::get_notify_fd does not move the
-        // ownership of the file descriptor so we must not close it.
         let seccomp_fd = match seccomp_filter.get_notify_fd() {
             Ok(fd) => {
-                // SAFETY: Duplicate the fd to a random location,
-                // for added hardening. This fd is O_CLOEXEC too.
-                match duprand(fd) {
-                    Ok(fd_rand) => {
-                        let _ = close(fd);
-                        fd_rand
-                    }
-                    Err(errno) => fail_errno_safe!(child, Err::Seccomp, errno as i32),
-                }
+                // SAFETY: get_notify_fd returns a valid FD.
+                unsafe { OwnedFd::from_raw_fd(fd) }
             }
             Err(scmp_err) => fail_errno_safe!(
                 child,
@@ -163,124 +176,64 @@ pub extern "C" fn child_after_clone(arg: *mut libc::c_void) -> libc::c_int {
             ),
         };
 
-        // Write the value of the fd to the pipe.
+        // Write the value of the seccomp notify fd to the pipe.
         // Handle partial writes and interrupts.
-        let fd_bytes = seccomp_fd.to_le_bytes();
+        // EOF means parent died before reading.
+        let fd = seccomp_fd.as_raw_fd().to_le_bytes();
         let mut nwrite = 0;
-        while nwrite < fd_bytes.len() {
+        while nwrite < fd.len() {
             #[allow(clippy::arithmetic_side_effects)]
-            match write(pipe_rw, &fd_bytes[nwrite..]) {
+            match retry_on_eintr(|| write(&pipe_rw, &fd[nwrite..])) {
                 Ok(0) => {
-                    // Parent died before reading,
-                    // this should never happen.
+                    // Parent died before reading.
+                    // This should ideally never happen.
                     fail_errno_safe!(child, Err::SeccompSendFd, Errno::EIO as i32);
                 }
                 Ok(n) => nwrite += n,
-                Err(Errno::EINTR | Errno::EAGAIN) => continue,
                 Err(errno) => fail_errno_safe!(child, Err::SeccompSendFd, errno as i32),
             }
         }
 
+        // Close the write end of the pipe.
+        drop(pipe_rw);
+
         // Wait for the parent to get the file descriptor.
-        // Read PID-FD number as reply from the pipe.
-        // Handle partial reads and interrupts.
-        let mut pid_buf = [0u8; std::mem::size_of::<RawFd>()];
-        let mut nread = 0;
-        while nread < pid_buf.len() {
-            #[allow(clippy::arithmetic_side_effects)]
-            match read(pipe_ro.as_raw_fd(), &mut pid_buf[nread..]) {
-                Ok(0) => {
-                    // Parent died before writing,
-                    // this should never happen.
-                    fail_errno_safe!(child, Err::SeccompWaitFd, Errno::EIO as i32);
-                }
-                Ok(n) => nread += n,
-                Err(Errno::EINTR | Errno::EAGAIN) => continue,
-                Err(errno) => fail_errno_safe!(child, Err::SeccompWaitFd, errno as i32),
+        // Handle interrupts.
+        // Partial read is not possible.
+        // EOF means parent died before writing to the pipe.
+        let mut buf = [0u8; 1];
+        match retry_on_eintr(|| read(&pipe_ro, &mut buf[..])) {
+            Ok(0) => {
+                // Parent died before writing.
+                // This should ideally never happen.
+                fail_errno_safe!(child, Err::SeccompWaitFd, Errno::EIO as i32);
             }
+            Ok(1) if buf[0] == 42 => {
+                // Parent received seccomp fd successfully.
+                // We can go ahead and close our copy now.
+            }
+            Ok(_) => unreachable!("BUG: The meaning of life is not {:#x}!", buf[0]),
+            Err(errno) => fail_errno_safe!(child, Err::SeccompWaitFd, errno as i32),
         }
 
-        let pid_fd = match pid_buf.as_slice().try_into() {
-            Ok(bytes) => RawFd::from_le_bytes(bytes),
-            Err(_) => fail_errno_safe!(child, Err::SeccompWaitFd, Errno::EINVAL as i32),
-        };
+        // Close our copy of the seccomp-notify fd.
+        // Parent process has already acknowledged that
+        // it has received a copy of this fd.
+        drop(seccomp_fd);
 
-        Some((seccomp_fd, pid_fd))
+        // Release resources for seccomp BPF filter.
+        // Memory allocation/deallocation is OK here
+        // now that we have transferred over the
+        // seccomp-notify fd to the parent process.
+        // Otherwise we'd risk breaking Memory sandboxing.
+        drop(seccomp_filter);
+
+        // Close the read end of the pipe.
+        drop(pipe_ro);
     } else {
-        None
-    };
-
-    // Drop the seccomp filter which is going to close down the owned
-    // seccomp notify file descriptor. We do this after the fd transfer
-    // is done to ensure a potential memory {de,}allocation system call
-    // cannot deadlock the process in case memory sandboxing is on.
-    drop(child.seccomp_filter);
-
-    // Close down the pipe now that the transfer is done.
-    // Single drop is adequate here, beauty of owned fds ;)
-    drop(child.seccomp_pipefd);
-
-    // SAFETY: Do not leak the log file descriptor to the sandbox process.
-    // We will close it after CLONE_FILES for library load,
-    // and set it to O_CLOEXEC for program runs.
-    let log_fd = if let Some(fd) = env::var_os(ENV_LOG_FD) {
-        btoi::<RawFd>(fd.as_bytes()).ok()
-    } else {
-        None
-    };
-
-    match child.exe {
-        Exe::Program(_) => {
-            // Nothing left to do, both seccomp-fd and pid-fd
-            // are close-on-exec and file table will be unshared
-            // on execve(2). Careful, the fd may be invalid...
-            if let Some(fd) = log_fd {
-                match set_cloexec(&fd, true) {
-                    Ok(_) | Err(Errno::EBADF) => {}
-                    Err(errno) => fail_errno_safe!(child, Err::UnshareFiles, errno as i32),
-                }
-            }
-        }
-        Exe::Library(_) => {
-            // Remove the effect of CLONE_FILES, and
-            // close down our copies of the pid-fd and seccomp-fd.
-            // We only need to do this if we're loading a library,
-            // as execve(2) already unshares the file table itself.
-            if let Err(errno) = unshare(CloneFlags::CLONE_FILES) {
-                fail_errno_safe!(child, Err::UnshareFiles, errno as i32);
-            }
-
-            // Close static file descriptors used by syd::proc and friends.
-            proc_close();
-
-            if let Some((seccomp_fd, pid_fd)) = fdpair {
-                let _ = close(pid_fd);
-                let _ = close(seccomp_fd);
-            }
-
-            // Close the log descriptor if any.
-            if let Some(fd) = log_fd {
-                let _ = close(fd);
-            }
-
-            // SAFETY: Do not leak the poll file descriptor to the sandbox process.
-            // Note, we only have to do this for library load,
-            // as epoll file descriptor is O_CLOEXEC and will be closed for program runs.
-            if let Some(fd) = env::var_os(ENV_POLL_FD) {
-                if let Ok(fd) = btoi::<RawFd>(fd.as_bytes()) {
-                    let _ = close(fd);
-                }
-            }
-
-            // SAFETY: Even after all this, some fds can leak into the library.
-            // See: https://builds.sr.ht/~alip/job/1409189#task-test for an example.
-            // Until a valid use-case is displayed, let's close all but the standard fds.
-            if let Err(errno) = Errno::result(unsafe {
-                libc::syscall(libc::SYS_close_range, 3, libc::c_int::MAX, 0)
-            }) {
-                fail_errno_safe!(child, Err::UnshareFiles, errno as i32);
-            }
-        }
+        // Close unused ends of the pipes.
+        drop(pipe_ro);
+        drop(pipe_rw);
     }
 
     // Drop CAP_SYS_PTRACE late as Syd may need it.
@@ -291,12 +244,13 @@ pub extern "C" fn child_after_clone(arg: *mut libc::c_void) -> libc::c_int {
         fail_safe!(child, Err::CapSet);
     }
 
-    // SAFETY: Clean Syd environment variables from process environment.
-    for (key, _) in env::vars_os() {
-        if key.as_bytes().starts_with(b"SYD_") && !key.as_bytes().starts_with(b"SYD_TEST_") {
-            env::remove_var(key);
-        }
-    }
+    // TODO:
+    // Set the new secure bits:
+    // 1. SECBIT_EXEC_RESTRICT_FILE
+    // 2. SECBIT_EXEC_DENY_INTERACTIVE
+    // upon user configuration.
+    // See: https://docs.kernel.org/userspace-api/check_exec.html
+    // Note, we already use AT_EXECVE_CHECK.
 
     match child.exe {
         Exe::Library(lib) => unsafe {
